/*
Copyright (C) 2020 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2017 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "CommandLineParsing.h"
#include "..\Tools\String\AnsiUnicodeConvert.h"

#ifdef _WIN64
    #define SOFTWARE_FILE_NAME    TEXT("DesktopIconsPositions64.exe")
#else
    #define SOFTWARE_FILE_NAME    TEXT("DesktopIconsPositions.exe")
#endif

#ifndef DefineWCharStringLength
    #define DefineWCharStringLengthWithEndingZero(x) (sizeof(x)/sizeof(WCHAR))
    #define DefineWCharStringLength(x) (sizeof(x)/sizeof(WCHAR)-1)
#endif

#define WCharAndSize(x) x,DefineWCharStringLength(x)

#define DEFAULT_CONFIG_FILE TEXT("DesktopIconsPositions.ini")

void ShowCommandLineHelp()
{
    ::MessageBox(NULL,
            SOFTWARE_FILE_NAME TEXT("save/restore [verbose]\r\n")
            SOFTWARE_FILE_NAME TEXT("[save/restore=]FilePath [verbose]\r\n")
            TEXT(" save : save icons positions to default file (") DEFAULT_CONFIG_FILE TEXT(")\r\n")
            TEXT(" restore : reloads icons positions from default file (") DEFAULT_CONFIG_FILE TEXT(")\r\n")
            TEXT(" save=FilePath : save icons positions to specified file (relative or absolute path)\r\n")
            TEXT(" restore=FilePath : reloads icons positions from specified file (relative or absolute path)\r\n")
            TEXT(" FilePath : reloads icons positions from specified file (relative or absolute path)\r\n")
            TEXT(" verbose : [Optional] display dialog message even on success")
            ,
            TEXT("Desktop Icon Position Command Line Help"),
            MB_OK|MB_ICONINFORMATION | MB_SYSTEMMODAL
            );
}

//-----------------------------------------------------------------------------
// Name: ParseCommandLine
// Object: parse cmd line
// Parameters :
//     BOOL* pbCommandLineMode : TRUE if command line has parameters
//     return : TRUE on SUCCESS, FALSE on error
//-----------------------------------------------------------------------------
BOOL ParseCommandLine(OUT BOOL* pbCommandLineMode,IN OUT CCommandLineInfo* pCommandLineInfo)
{
    BOOL bRes=TRUE;
    LPWSTR* lppwargv;
    int argc=0;
    int cnt;

    *pbCommandLineMode = FALSE;
    
    // use CommandLineToArgvW
    // Notice : CommandLineToArgvW translate 'option="a";"b"' into 'option=a;b'
    lppwargv=CommandLineToArgvW(
                                GetCommandLineW(),
                                &argc
                                );
    // if no params
    if (argc<=1) // argc == 1 --> only app name aka no args
    {
        LocalFree(lppwargv);
        return TRUE;
    }

    // we are in command line mode
    *pbCommandLineMode=TRUE;

    // for each param
    // for each param
    for(cnt=0;cnt<argc;cnt++)
    {
        // depending the way CreateProcess is called, param 0 can be the name of the application or not
        if (cnt == 0)
        {
            WCHAR AppPathW[MAX_PATH];
            ::GetModuleFileNameW(NULL,AppPathW,_countof(AppPathW));
            if (_wcsicmp(lppwargv[0],AppPathW)==0)
                continue;
        }

        if ( (_wcsicmp(lppwargv[cnt],L"--help")==0)
            || (_wcsicmp(lppwargv[cnt],L"/Help")==0)
            || (_wcsicmp(lppwargv[cnt],L"Help")==0)
            || (_wcsicmp(lppwargv[cnt],L"-H")==0)
            || (_wcsicmp(lppwargv[cnt],L"-?")==0)
            || (_wcsicmp(lppwargv[cnt],L"/H")==0)
            || (_wcsicmp(lppwargv[cnt],L"/?")==0)
            || (_wcsicmp(lppwargv[cnt],L"?")==0)
            )
        {
            // show help message
            ShowCommandLineHelp();
            return FALSE;
        }
        else if (_wcsicmp(lppwargv[cnt],L"verbose")==0)
        {
            pCommandLineInfo->bVerbose=TRUE;
        }
        else if ( _wcsnicmp( lppwargv[cnt],WCharAndSize(L"save=") )==0)
        {
            pCommandLineInfo->Operation = CCommandLineInfo::Operation_SAVE;
            WCHAR* wc = lppwargv[cnt]+DefineWCharStringLength(L"save=");
            pCommandLineInfo->FreeFilePath();
            CAnsiUnicodeConvert::UnicodeToTchar(wc,&pCommandLineInfo->pFilePath);
        }
        else if ( _wcsnicmp( lppwargv[cnt],WCharAndSize(L"restore=") )==0)
        {
            pCommandLineInfo->Operation = CCommandLineInfo::Operation_RESTORE;
            WCHAR* wc = lppwargv[cnt]+DefineWCharStringLength(L"restore=");
            pCommandLineInfo->FreeFilePath();
            CAnsiUnicodeConvert::UnicodeToTchar(wc,&pCommandLineInfo->pFilePath);
        }
        else if (_wcsicmp(lppwargv[cnt],L"save")==0)
        {
            pCommandLineInfo->Operation = CCommandLineInfo::Operation_SAVE;
            pCommandLineInfo->FreeFilePath();
            pCommandLineInfo->pFilePath = _tcsdup(DEFAULT_CONFIG_FILE);
        }
        else if (_wcsicmp(lppwargv[cnt],L"restore")==0)
        {
            pCommandLineInfo->Operation = CCommandLineInfo::Operation_RESTORE;
            pCommandLineInfo->FreeFilePath();
            pCommandLineInfo->pFilePath = _tcsdup(DEFAULT_CONFIG_FILE);
        }                       
        else // arg without prefix : supposed to be file path
        {
            if (pCommandLineInfo->pFilePath == NULL)
            {
                CAnsiUnicodeConvert::UnicodeToTchar(lppwargv[cnt],&pCommandLineInfo->pFilePath);
            }
            else
            {
                ShowCommandLineHelp();
                return FALSE;
            }

            if (pCommandLineInfo->Operation == CCommandLineInfo::Operation_NONE)
            {
                pCommandLineInfo->Operation = CCommandLineInfo::Operation_RESTORE;
            }
        }
    }
    LocalFree(lppwargv);

    return bRes;
}