/*
Copyright (C) 2020 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2017 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "DialogDesktopIconsPositions.h"
#include <CommCtrl.h>

#include "../Tools/GUI/DesktopWorkArea/DesktopWorkArea.h"
#include "../Tools/Ini/IniFile.h"
#include "../Tools/Donations/Donations.h"
#include "../Tools/Version/Version.h"
#include "../Tools/SoftwareUpdate/SoftwareUpdate.h"

#include "resource.h"
#include "about.h"

#define CDialogDesktopIconsPositions_ThumbnailVwExtWnd32 _T("ThumbnailVwExtWnd32")
#define CDialogDesktopIconsPositions_ITEM_POSITION_ERROR  -1

CDialogDesktopIconsPositions::CDialogDesktopIconsPositions()
{
    ::InitializeCriticalSection(&this->RealDesktopItemsInfosCriticalSection);
    
    this->pToolbar = NULL;
    this->hDesktopListView = NULL;
    this->IconWidth  = ::GetSystemMetrics(SM_CXICON);
    this->IconHeight = ::GetSystemMetrics(SM_CYICON);
}

CDialogDesktopIconsPositions::~CDialogDesktopIconsPositions()
{
    ::DeleteCriticalSection(&this->RealDesktopItemsInfosCriticalSection);
}

BOOL CDialogDesktopIconsPositions::OnInit()
{
    this->pToolbar = new CToolbar(this->hInstance,this->hWndDialog,TRUE,TRUE,24,24,0);
    //this->pToolbar->SetEventsHandler((IToolbarEvents*)(CToolbarEventsBase*)this);
    this->pToolbar->AddButton(IDC_BUTTON_SAVE,IDI_ICON_SAVE,TEXT("Save Desktop Icons Positions"),24,24);
    this->pToolbar->AddButton(IDC_BUTTON_RESTORE,IDI_ICON_OPEN,TEXT("Restore Desktop Icons Positions"),24,24);
    this->pToolbar->AddSeparator();
    this->pToolbar->AddCheckButton(IDC_BUTTON_MULTIPLE_CONFIGS,IDI_ICON_MULTIPLE_CONFIGS,TEXT("Allow to manage multiple configurations"),24,24);
    this->pToolbar->AddSeparator();
    this->pToolbar->AddButton(IDC_BUTTON_CHECK_FOR_UPDATE,IDI_ICON_CHECK_FOR_UPDATE,TEXT("Check for update"),24,24);
    this->pToolbar->AddButton(IDC_BUTTON_BUG_REPORT,IDI_ICON_BUG_REPORT,TEXT("Report Bug"),24,24);
    this->pToolbar->AddButton(IDC_BUTTON_ABOUT,IDI_ICON_ABOUT,TEXT("About"),24,24);
    this->pToolbar->AddSeparator();
    this->pToolbar->AddButton(IDC_BUTTON_DONATION,IDI_ICON_DONATION,TEXT("Make Donation"),24,24);

    this->SetDlgButtonIcon(IDC_BUTTON_SAVE,IDI_ICON_SAVE,64,64);
    this->SetDlgButtonIcon(IDC_BUTTON_RESTORE,IDI_ICON_OPEN,64,64);


    DWORD dw;
    TCHAR IniFilePath[MAX_PATH];
    this->GetIniFilePath(IniFilePath,_countof(IniFilePath));
    CIniFile IniFile(IniFilePath);
    IniFile.Get(TEXT("MultipleConfigs"),0,&dw);
    if (dw!=0)
        this->pToolbar->SetButtonState(IDC_BUTTON_MULTIPLE_CONFIGS,TBSTATE_ENABLED|TBSTATE_CHECKED);

#ifndef _WIN64
    // if 32 bits application running on 64 bits OS
    SYSTEM_INFO SystemInfo={0};
    ::GetNativeSystemInfo(&SystemInfo);
    switch(SystemInfo.wProcessorArchitecture)
    {
    case PROCESSOR_ARCHITECTURE_AMD64:    // 9    x64 (AMD or Intel)
        // as we need to send messages to explorer process, structs of messages sent MUST BE of the same size as the explorer process
        // ans as size of structs are different in 32 and 64 bits process we have to put limitation

        // disable save and restore as they could crash application displaying desktop
        this->pToolbar->EnableButton(IDC_BUTTON_SAVE,FALSE);
        this->pToolbar->EnableButton(IDC_BUTTON_RESTORE,FALSE);
        this->EnableDlgItem(IDC_BUTTON_SAVE,FALSE);
        this->EnableDlgItem(IDC_BUTTON_RESTORE,FALSE);

        // report bad usage
        this->ReportError(CDialogDesktopIconsPositions_PROJECT_NAME TEXT(" 32 bit version must be run only on 32 bit operating system.\r\nFor your OS please use the 64 bits version of ") CDialogDesktopIconsPositions_PROJECT_NAME);

        // display web page with 64bit version
        ::ShellExecute(this->hWndDialog,_T("open"),CDialogDesktopIconsPositions_WEB_PROJECT_LINK,NULL,NULL,SW_SHOWNORMAL);

        break;
    }
#endif

    return TRUE;
}

BOOL CDialogDesktopIconsPositions::OnClose()
{
    TCHAR IniFilePath[MAX_PATH];
    this->GetIniFilePath(IniFilePath,_countof(IniFilePath));
    CIniFile IniFile(IniFilePath);
    IniFile.Set(TEXT("MultipleConfigs"),(DWORD)this->pToolbar->IsButtonChecked(IDC_BUTTON_MULTIPLE_CONFIGS));

    DeleteAndCleanIfNotNull(this->pToolbar);
    return TRUE;
}

void CDialogDesktopIconsPositions::OnCommand(WPARAM wParam,LPARAM lParam)
{
    UNREFERENCED_PARAMETER(lParam);
    switch(LOWORD(wParam))
    {
    case IDC_BUTTON_SAVE:
        this->OnSave();
        break;
    case IDC_BUTTON_RESTORE:
        this->OnRestore();
        break;
    case IDCANCEL:
        this->Close(IDCANCEL);
        break;

    case IDC_BUTTON_CHECK_FOR_UPDATE:
        this->CheckForUpdate();
        break;
    case IDC_BUTTON_BUG_REPORT:
        this->BugReport();
        break;
    case IDC_BUTTON_ABOUT:
        {
            CAbout About;
            About.Show(this->hInstance,this->hWndDialog,IDD_DIALOG_ABOUT,IDI_ICON_ABOUT);
        }
        break;
    case IDC_BUTTON_DONATION:
        CDonations::MakePaypalDonation(this->hWndDialog,
                                        CDialogDesktopIconsPositions_PROJECT_NAME,
                                        CDialogDesktopIconsPositions_WEB_PROJECT_LINK
                                        );
        break;

    }
}

void CDialogDesktopIconsPositions::CheckForUpdate()
{
    CSoftwareUpdate::CheckForUpdate(
                                    CDialogDesktopIconsPositions_PAD_Url,
                                    CDialogDesktopIconsPositions_VersionSignificantDigits,
                                    CDialogDesktopIconsPositions_DownloadLink,
                                    TRUE,
                                    NULL,  // pbNewVersionIsAvailable,
                                    FALSE, // bShowOnlyNewVersionMsg
                                    this->hWndDialog
                                    );
}

void CDialogDesktopIconsPositions::BugReport()
{
    TCHAR SoftwareFullPath[MAX_PATH];
    TCHAR PrettyVersion[64];
    TCHAR Subject[256];
    CStdFileOperations::GetAppPath(SoftwareFullPath,_countof(SoftwareFullPath));
    CVersion Version;
    Version.Read(SoftwareFullPath);
    *PrettyVersion=0;
    CVersion::GetPrettyVersion(Version.ProductVersion,3,PrettyVersion,_countof(PrettyVersion));

    // see RFC 2368 for mailto protocol
    TCHAR pszURL[2048];
    _tcscpy(pszURL,_T("mailto:jacquelin.potier@free.fr?subject="));
    _sntprintf(Subject,_countof(Subject), CDialogDesktopIconsPositions_PROJECT_NAME _T(" %s Bug Report"),PrettyVersion);
    _tcscat(pszURL,Subject);
    _tcscat(pszURL,
            _T("&body=Please provide the most possible details about the bug, ")
            _T("including the description of the problem itself.%0D%0A")
            _T("Please include any error messages you received.%0D%0A")
            _T("All these details will be useful for a quicker bug correction%0D%0A%0D%0A")
            _T("Thanks for the report%0D%0A")
            _T("Jacquelin")
            );
    if (::ShellExecute(NULL,_T("open"),pszURL,NULL,NULL,SW_SHOWNORMAL)<(HINSTANCE)33)
    {
        this->ReportError(_T("Error launching default mail client"));
    }
}

BOOL CDialogDesktopIconsPositions::OnNotify(WPARAM wParam,LPARAM lParam)
{
    if (this->pToolbar)
    {
        if (this->pToolbar->OnNotify(wParam,lParam))
            return TRUE;
    }

    return FALSE;
}

void CDialogDesktopIconsPositions::GetIniFilePath(TCHAR* IniFilePath,DWORD IniFilePathMaxSize)
{
    TCHAR ApplicationDirectory[MAX_PATH];
    CStdFileOperations::GetAppDirectory(ApplicationDirectory,_countof(ApplicationDirectory));

    // single configuration file
    _sntprintf(IniFilePath,IniFilePathMaxSize,
                _T("%sDesktopIconsPositions.ini"),
                ApplicationDirectory
                );
    IniFilePath[IniFilePathMaxSize-1]=0;
}

BOOL CDialogDesktopIconsPositions::OnSave()
{
    // get ini file name
    TCHAR ApplicationDirectory[MAX_PATH];
    CStdFileOperations::GetAppDirectory(ApplicationDirectory,_countof(ApplicationDirectory));
    TCHAR SavingFileName[MAX_PATH];

    BOOL bMulitpleConfigs = this->pToolbar->IsButtonChecked(IDC_BUTTON_MULTIPLE_CONFIGS);
    if (bMulitpleConfigs)
    {
        SYSTEMTIME Time;
        ::GetLocalTime(&Time);
        _sntprintf(SavingFileName,_countof(SavingFileName),
                    _T("%sDesktopIconsPositions_%04u_%02u_%02u_%02u_%02u_%02u.ini"),
                    ApplicationDirectory,
                    Time.wYear,
                    Time.wMonth,
                    Time.wDay,
                    Time.wHour,
                    Time.wMinute,
                    Time.wSecond
                    );

        // open dialog
        OPENFILENAME ofn;
        memset(&ofn,0,sizeof (OPENFILENAME));
        ofn.lStructSize=sizeof (OPENFILENAME);
        ofn.hwndOwner=this->hWndDialog;
        ofn.hInstance=this->hInstance;
        ofn.lpstrFilter=TEXT("ini\0*.ini\0");
        ofn.nFilterIndex = 1;
        ofn.Flags=OFN_EXPLORER|OFN_NOREADONLYRETURN|OFN_OVERWRITEPROMPT;
        ofn.lpstrDefExt=TEXT("ini");
        ofn.lpstrFile=SavingFileName;
        ofn.nMaxFile=MAX_PATH;

        if (!GetSaveFileName(&ofn))
            return FALSE;

    }
    else
    {
        // single configuration file
        _sntprintf(SavingFileName,_countof(SavingFileName),
                    _T("%sDesktopIconsPositions.ini"),
                    ApplicationDirectory
                    );
    }

    return this->Save(SavingFileName,TRUE);
}

BOOL CDialogDesktopIconsPositions::Save(TCHAR* AbsoluteOrRelativePath,BOOL bMsgBoxOnSuccess)
{
    // get desktop icons positions
    if (!this->GetDesktopItemsPos())
        return FALSE;

    TCHAR BackupFileName[MAX_PATH];
    TCHAR* SavingFileName = AbsoluteOrRelativePath;
    TCHAR AbsolutePath[MAX_PATH];
    if (CStdFileOperations::IsRelativePath(AbsoluteOrRelativePath))
    {
        if (CStdFileOperations::GetAbsolutePath(AbsoluteOrRelativePath,AbsolutePath) == FALSE )
        {
            TCHAR Msg[512];
            _sntprintf(Msg,_countof(Msg),TEXT("Error resolving full path for :%s"),AbsoluteOrRelativePath);
            Msg[_countof(Msg)-1]=0;
            this->ReportError(Msg);
            return FALSE;
        }
        SavingFileName = AbsolutePath;
    }

    // generate backup file name
    _sntprintf(BackupFileName,_countof(BackupFileName),TEXT("%s.bak"),SavingFileName);
    BackupFileName[_countof(BackupFileName)-1]=0;
    

    //if (CStdFileOperations::DoesFileExists(BackupFileName))
    //    ::DeleteFile(BackupFileName);

    if (CStdFileOperations::DoesFileExists(SavingFileName))
    {
        if (!::MoveFileEx(SavingFileName,BackupFileName,MOVEFILE_REPLACE_EXISTING | MOVEFILE_WRITE_THROUGH))
        {
            TCHAR Msg[512];
            _sntprintf(Msg,_countof(Msg),TEXT("Operation cancelled: error creating backup file :%s"),BackupFileName);
            Msg[_countof(Msg)-1]=0;
            this->ReportError(Msg);
            return FALSE;
        }
    }

    CIniFile IniFile(SavingFileName);
    if (!IniFile.IsFileOpenSuccessfull())
    {
        TCHAR Msg[512];
        _sntprintf(Msg,_countof(Msg),TEXT("Error opening file :%s"),SavingFileName);
        Msg[_countof(Msg)-1]=0;
        this->ReportError(Msg);
        return FALSE;
    }

    // store icon position and labels into .ini
    std::vector<REAL_DESKTOP_ITEM_INFOS>::iterator RealDesktopItemInfosIterator;

    this->SetWaitCursor(TRUE);
    ::EnterCriticalSection(&this->RealDesktopItemsInfosCriticalSection);

    DWORD Cnt=0;
    TCHAR SectionName[128];

    BOOL bTmpRet;
    BOOL bRet = TRUE;

    // for each desktop icon
    for (RealDesktopItemInfosIterator = this->RealDesktopItemsInfos.begin();
        RealDesktopItemInfosIterator!=this->RealDesktopItemsInfos.end(); 
        RealDesktopItemInfosIterator++
        )
    {
        // as we are referencing by name, if name is empty no way to restore --> no interest to save
        if (*RealDesktopItemInfosIterator->Name == 0)
            continue;

        _sntprintf(SectionName,_countof(SectionName),TEXT("Icon_%u"),Cnt);
        IniFile.SetCurrentSectionName(SectionName);
        bTmpRet = IniFile.Set(TEXT("Name"),RealDesktopItemInfosIterator->Name);
        bRet = bRet && bTmpRet;
        bTmpRet = IniFile.Set(TEXT("PosX"),(DWORD)RealDesktopItemInfosIterator->Position.x);
        bRet = bRet && bTmpRet;
        bTmpRet = IniFile.Set(TEXT("PosY"),(DWORD)RealDesktopItemInfosIterator->Position.y);
        bRet = bRet && bTmpRet;

        Cnt++;
    }

    IniFile.SetCurrentSectionName(TEXT("Default"));
    IniFile.Set(TEXT("NbIcons"),Cnt);

    ::LeaveCriticalSection(&this->RealDesktopItemsInfosCriticalSection);
    this->SetWaitCursor(FALSE);

    if (bRet)
    {
        if (bMsgBoxOnSuccess)
            this->ReportMessage(TEXT("Desktop icons positions successfully saved"),ReportMessageType_INFO);
    }
    else
        this->ReportMessage(TEXT("Error saving desktop icons positions"),ReportMessageType_ERROR);

    return bRet;
}

BOOL CDialogDesktopIconsPositions::OnRestore()
{
    // get ini file name
    TCHAR ApplicationDirectory[MAX_PATH];
    CStdFileOperations::GetAppDirectory(ApplicationDirectory,_countof(ApplicationDirectory));
    TCHAR RestoreFileName[MAX_PATH];
    _sntprintf(RestoreFileName,_countof(RestoreFileName),
        TEXT("%sDesktopIconsPositions.ini"),
        ApplicationDirectory
        );

    BOOL bMulitpleConfigs = this->pToolbar->IsButtonChecked(IDC_BUTTON_MULTIPLE_CONFIGS);
    if (bMulitpleConfigs)
    {
        // open dialog
        OPENFILENAME ofn;
        memset(&ofn,0,sizeof (OPENFILENAME));
        ofn.lStructSize=sizeof (OPENFILENAME);
        ofn.hwndOwner=this->hWndDialog;
        ofn.hInstance=this->hInstance;
        ofn.lpstrFilter=TEXT("ini\0*.ini\0");
        ofn.nFilterIndex = 1;
        ofn.Flags=OFN_EXPLORER|OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST;
        ofn.lpstrDefExt=TEXT("ini");
        ofn.lpstrFile=RestoreFileName;
        ofn.nMaxFile=MAX_PATH;

        if (!GetOpenFileName(&ofn))
            return FALSE;
    }

    return this->Restore(RestoreFileName,TRUE);
}

BOOL CDialogDesktopIconsPositions::PointBelongsToVisiblePartOfMonitor(POINT* pPoint)
{
    DWORD MonitorIndex;
    RECT MonitorRect;
    return CDesktopWorkArea::GetWorkAreaForMonitorFromPoint(pPoint,&MonitorIndex,&MonitorRect,FALSE);
}


BOOL CDialogDesktopIconsPositions::Restore(TCHAR* AbsoluteOrRelativePath,BOOL bMsgBoxOnSuccess)
{
    // get current icon position to get labels
    if (!this->GetDesktopItemsPos())
        return FALSE;

    // get absolute config file path
    TCHAR* RestoreFileName = AbsoluteOrRelativePath;
    TCHAR AbsolutePath[MAX_PATH];
    if (CStdFileOperations::IsRelativePath(AbsoluteOrRelativePath))
    {
        if (CStdFileOperations::GetAbsolutePath(AbsoluteOrRelativePath,AbsolutePath) == FALSE )
        {
            TCHAR Msg[512];
            _sntprintf(Msg,_countof(Msg),TEXT("Error resolving full path for :%s"),AbsoluteOrRelativePath);
            Msg[_countof(Msg)-1]=0;
            this->ReportError(Msg);
            return FALSE;
        }
        RestoreFileName = AbsolutePath;
    }
    
    // check if config file exists
    if (!CStdFileOperations::DoesFileExists(RestoreFileName))
    {
        TCHAR Msg[512];
        _sntprintf(Msg,_countof(Msg),TEXT("File %s doesn't exists"),RestoreFileName);
        Msg[_countof(Msg)-1]=0;
        this->ReportError(Msg);
        return FALSE;
    }

    // open config file
    CIniFile IniFile(RestoreFileName);
    if (!IniFile.IsFileOpenSuccessfull())
    {
        TCHAR Msg[512];
        _sntprintf(Msg,_countof(Msg),TEXT("Error opening file :%s"),RestoreFileName);
        Msg[_countof(Msg)-1]=0;
        this->ReportError(Msg);
        return FALSE;
    }

    // get number of icons to restore
    DWORD NbIconsToRestore;
    IniFile.SetCurrentSectionName(TEXT("Default"));
    IniFile.Get(TEXT("NbIcons"),0,&NbIconsToRestore);
    if (NbIconsToRestore == 0)
    {
        this->ReportError(TEXT("Empty configuration: No icon to restore"));
        return FALSE;
    }

    // store icon position and labels into .ini
    std::vector<REAL_DESKTOP_ITEM_INFOS>::iterator RealDesktopItemInfosIterator;

    this->SetWaitCursor(TRUE);
    ::EnterCriticalSection(&this->RealDesktopItemsInfosCriticalSection);

    DWORD Cnt;
    TCHAR SectionName[128];
    TCHAR Name[MAX_PATH];
    POINT Position;

    BOOL bValidPosition;
    BOOL bSetPosition;
    BOOL bRealDesktopItemFound;

    // for each icon to restore
    for(Cnt=0;Cnt<NbIconsToRestore;Cnt++)
    {
        bSetPosition = TRUE;
        bRealDesktopItemFound = FALSE;

        // load icon position and labels from .ini
        _sntprintf(SectionName,_countof(SectionName),TEXT("Icon_%u"),Cnt);
        IniFile.SetCurrentSectionName(SectionName);
        IniFile.Get(TEXT("Name"),TEXT(""),Name,MAX_PATH);
        IniFile.Get(TEXT("PosX"),0,(DWORD*)&Position.x);
        IniFile.Get(TEXT("PosY"),0,(DWORD*)&Position.y);

        // if icon to restore has no name
        if (*Name==0)
            continue;

        // for each desktop icon
        for (RealDesktopItemInfosIterator = this->RealDesktopItemsInfos.begin();
            RealDesktopItemInfosIterator!=this->RealDesktopItemsInfos.end(); 
            RealDesktopItemInfosIterator++
            )
        {
            // if real desktop icon has no name
            if (*RealDesktopItemInfosIterator->Name == 0)
                continue;

            // if the name of icon to restore match the real name
            if (_tcsicmp(RealDesktopItemInfosIterator->Name,Name)==0)
            {
                bRealDesktopItemFound = TRUE;
                break;
            }
        }
        if (bRealDesktopItemFound == FALSE)
            continue;
        
        // v1.2 : ensure current icon position is in a visible area (avoid to restore icon inside a removed monitor)
        bValidPosition = CDialogDesktopIconsPositions::PointBelongsToVisiblePartOfMonitor(&Position);

        // if the saved icon position does not belong to the visible part
        if (bValidPosition == FALSE)
        {
            // if current icon position belongs to visible part, let it as is
            // else put icon at a Default Position
            if (CDialogDesktopIconsPositions::PointBelongsToVisiblePartOfMonitor(&RealDesktopItemInfosIterator->Position))
                bSetPosition = FALSE;
            else
            {
                Position = this->PointDefaultPosition;

                // prepare next default position
                this->PointDefaultPosition.y+=this->IconHeight;

                if (!CDialogDesktopIconsPositions::PointBelongsToVisiblePartOfMonitor(&this->PointDefaultPosition))
                {
                    RECT MainWorkArea;
                    CDesktopWorkArea::GetWorkArea(&MainWorkArea,TRUE,FALSE);

                    this->PointDefaultPosition.y  = MainWorkArea.top;
                    this->PointDefaultPosition.x += this->IconWidth;

                    if (!CDialogDesktopIconsPositions::PointBelongsToVisiblePartOfMonitor(&this->PointDefaultPosition))
                        this->PointDefaultPosition.x = MainWorkArea.left;
                }
            }
        }

        if (bSetPosition)
        {
            // set desktop icon position
            ::SendMessage(this->hDesktopListView, LVM_SETITEMPOSITION, RealDesktopItemInfosIterator->ItemIndex, MAKELPARAM(Position.x, Position.y));
        }
    }

    ::LeaveCriticalSection(&this->RealDesktopItemsInfosCriticalSection);
    this->SetWaitCursor(FALSE);

    if (bMsgBoxOnSuccess)
        this->ReportMessage(TEXT("Desktop icons positions successfully restored"),ReportMessageType_INFO);

    return TRUE;
}

HWND CDialogDesktopIconsPositions::GetDesktopListView()
{
    if (this->hDesktopListView)
        return this->hDesktopListView;

    HWND hListView = NULL;

    // find well known window containing desktop
    HWND hWnd = ::FindWindow(_T("Progman"), _T("Program Manager"));//::FindWindow(_T("ProgMan"), NULL);

    // find shell view
    HWND hShellView = ::FindWindowEx(hWnd,NULL,_T("SHELLDLL_DefView"),NULL);

    // bug solved QuickWayToFolder 1.2.8
    // in case of use of direct X wallpaper _T("Progman") doesn't contain _T("SHELLDLL_DefView")
    if (hShellView==0)
    {
        // use of DirectX --> ::FindWindow(_T("WorkerW"), NULL); // trouble : multiple WorkerW
        HWND hWndDesktop = ::GetDesktopWindow();
        HWND hWndChild = 0;
        for(;;)
        {
            hWndChild = ::FindWindowEx( hWndDesktop, hWndChild,  _T("WorkerW"),  NULL );
            if (!hWndChild)
                goto CleanUp;
            hShellView = ::FindWindowEx(hWndChild,NULL,_T("SHELLDLL_DefView"),NULL);

            if (hShellView)
                break;
        }
    }

    // find associated list view control
    hListView = ::FindWindowEx(hShellView,NULL,WC_LISTVIEW,NULL);
    if (!::IsWindowVisible(hListView))
    {
        hListView = ::FindWindowEx(hShellView,NULL,CDialogDesktopIconsPositions_ThumbnailVwExtWnd32,NULL);
        hListView = ::FindWindowEx(hListView,NULL,WC_LISTVIEW,NULL);
    }

CleanUp:
    this->hDesktopListView = hListView;

    return this->hDesktopListView;
}

//-----------------------------------------------------------------------------
// Name: GetDesktopItemsPos
// Object: get real desktop items positions
//		/!\ Warning works only on correct identical OS architecture version LVITEM is platform dependent (32 and 64 bit structs are different)
// Parameters :
//     in  :
//     out :
//     return : TRUE on success, FALSE on error
//-----------------------------------------------------------------------------
BOOL CDialogDesktopIconsPositions::GetDesktopItemsPos()
{
    BOOL bRet=FALSE;
    HANDLE hProcess=NULL;
    REAL_DESKTOP_ITEM_INFOS* lpData = NULL;

    this->SetWaitCursor(TRUE);
    ::EnterCriticalSection(&this->RealDesktopItemsInfosCriticalSection);
    this->RealDesktopItemsInfos.clear();

    HWND hListView = this->GetDesktopListView();

    DWORD dwProcessId;
    DWORD dwAccess;

    // get process associated to desktop
    ::GetWindowThreadProcessId(hListView, &dwProcessId);

    // open process with allocation, read and write memory access
    dwAccess = PROCESS_VM_OPERATION|PROCESS_VM_READ|PROCESS_VM_WRITE;
    hProcess = ::OpenProcess(dwAccess, FALSE, dwProcessId);
    if(!hProcess)
        goto CleanUp;

    // allocate required structs in remote process memory
    lpData =(REAL_DESKTOP_ITEM_INFOS*)::VirtualAllocEx(hProcess, NULL, sizeof(REAL_DESKTOP_ITEM_INFOS)+sizeof(LVITEM), MEM_COMMIT, PAGE_READWRITE);
    if(!lpData)
        goto CleanUp;

    // get pointer to 2nd struct
    LVITEM* lpItem=(LVITEM*) ( ((PBYTE)lpData) + sizeof(REAL_DESKTOP_ITEM_INFOS) );

    REAL_DESKTOP_ITEM_INFOS ItemInfos;

    // prepare an LVITEM struct
    LVITEM lvi;
    memset(&lvi, 0, sizeof(LVITEM));
    lvi.mask = LVIF_TEXT;
    lvi.pszText	= (LPTSTR) lpData->Name;
    lvi.cchTextMax = MAX_PATH;

    // get number of item on desktop
    int ItemCount =(int) ::SendMessage(hListView, LVM_GETITEMCOUNT, 0, 0);

    RECT PrimaryDesktopWorkArea;
    RECT AllDisplayWorkArea;
    CDesktopWorkArea::GetWorkAreaEx(&PrimaryDesktopWorkArea,&AllDisplayWorkArea);
    // put primary rect into relative coordinate of top left display
    ::OffsetRect(&PrimaryDesktopWorkArea,-AllDisplayWorkArea.left,-AllDisplayWorkArea.top);
    ::OffsetRect(&AllDisplayWorkArea,-AllDisplayWorkArea.left,-AllDisplayWorkArea.top);

    // for each item
    for(int Cnt = 0; Cnt < ItemCount; Cnt++)
    {
        // get item pos
        ItemInfos.Position.x=CDialogDesktopIconsPositions_ITEM_POSITION_ERROR;
        ItemInfos.Position.y=CDialogDesktopIconsPositions_ITEM_POSITION_ERROR;
        if (::SendMessage(hListView, LVM_GETITEMPOSITION, Cnt, (LPARAM) &lpData->Position))
        {
            // translate result from remote process memory to current process memory
            ::ReadProcessMemory(hProcess, &lpData->Position, &ItemInfos.Position, sizeof(POINT), NULL);
        }

        // get item text
        lvi.iItem = Cnt;
        *ItemInfos.Name=0;
        ItemInfos.ItemIndex = Cnt;
        // update remote LVITEM struct
        ::WriteProcessMemory(hProcess, lpItem, &lvi, sizeof(LVITEM), NULL);
        if (::SendMessage(hListView, LVM_GETITEMTEXT, Cnt, (LPARAM) lpItem) > 0)
        {
            // translate result from remote process memory to current process memory
            ::ReadProcessMemory(hProcess, &lpData->Name, &ItemInfos.Name, MAX_PATH*sizeof(TCHAR), NULL);
        }

        // store item text and position into RealDesktopItemsInfos vector
        this->RealDesktopItemsInfos.push_back(ItemInfos);
    }

    bRet=TRUE;

CleanUp:
    ::LeaveCriticalSection(&this->RealDesktopItemsInfosCriticalSection);
    this->SetWaitCursor(FALSE);

    // free memory 

    if (lpData)
        ::VirtualFreeEx(hProcess, lpData, 0, MEM_RELEASE);
    if (hProcess)
        ::CloseHandle(hProcess);

    return bRet;
}
