/*
Copyright (C) 2004-2020 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2004-2013 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "DialogBase.h"

CDialogBase::CDialogBase(void)
{
    this->hWndDialog=NULL;
    this->hWndParentDialog=NULL;
    this->hInstance=NULL;
    this->bInsideDialogBoxFunction=FALSE;
    this->IconResourceId=NULL;
    this->hWaitCursor=NULL;
    this->bWaitCursor=FALSE;
    this->bDragAcceptFileRequired=FALSE;
    this->bDragAcceptFileStatusEnabled=FALSE;
    this->bInitDone=FALSE;
    this->bModalDialog = TRUE;
    this->DisableClosingCount=0;
    this->bCloseModalSubDialogsIfAny = FALSE;

    ::InitializeCriticalSection(&this->CriticalSection);
}

void CDialogBase::BringToTop()
{
    if (::IsIconic(this->hWndDialog))
        ::ShowWindow(this->hWndDialog,SW_RESTORE);
    
    ::SetWindowPos(this->hWndDialog, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE|SWP_SHOWWINDOW);
    ::SetWindowPos(this->hWndDialog, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE|SWP_SHOWWINDOW);

    ::BringWindowToTop(this->hWndDialog);
}

HWND CDialogBase::GetParentPopUpDialog()
{
    return CDialogBase::GetParentPopUpDialog(this->hWndDialog);
}


//https://groups.google.com/a/chromium.org/forum/#!topic/chromium-dev/Hirr_DkuZdw
//
//I've performed an experiment to test the behavior on Windows 7.  DialogA creates DialogB which then creates DialogC.
//
//DialogA = 60f3e
//GetParent(DialogA) = 0
//GetAncestor(DialogA, GA_ROOT) = 60f3e
//GetAncestor(DialogA, GA_ROOTOWNER) = 60f3e
//DialogB = 30fde
//GetParent(DialogB) =60f3e
//GetAncestor(DialogB, GA_ROOT) = 30fde
//GetAncestor(DialogB, GA_ROOTOWNER) = 60f3e
//DialogC = 30fea
//GetParent(DialogC) =30fde
//GetAncestor(DialogC, GA_ROOT) = 30fea
//GetAncestor(DialogC, GA_ROOTOWNER) = 60f3e
//
//I also checked the Wine implementation of GetAncestor() and GetParent():
//http://source.winehq.org/source/dlls/user32/win.c#L2609
//http://source.winehq.org/source/dlls/user32/win.c#L2568
//
//If the window passed to GetParent() has the WS_POPUP flag it returns the window's owner and if it has the WS_CHILD flag it returns the window's parent.  This matches with the experimental results.
//
//All of this indicates the following:
//
//1. GetAncestor(hwnd, GA_ROOTOWNER) walks the child hierarchy using GetParent() to reach the top-level window.
//2. GetAncestor(hwnd, GA_ROOT) returns the current dialog window.
//
//So, in conclusion, if the intent is to retrieve the top-level window we should be using GetAncestor(hwnd, GA_ROOTOWNER) and if the intent is to return the current dialog window we should be using GetAncestor(hwnd, GA_ROOT).

HWND CDialogBase::GetParentTopMostPopUpDialog(HWND hWnd)
{
    return ::GetAncestor(hWnd, GA_ROOTOWNER);
}

HWND CDialogBase::GetParentPopUpDialog(HWND hWnd)
{
    return ::GetAncestor(hWnd, GA_ROOT);

    //HWND hWndParent=hWnd ;
    //
    //for(;;)
    //{
    //    hWndParent = (HWND)::GetWindowLongPtr(hWndParent ,GWLP_HWNDPARENT); 
    //    if (hWndParent == NULL)
    //        return NULL;
    //    // to check if required
    //    if ( ::GetWindowLongPtr(hWndParent ,GWL_STYLE) & WS_POPUP )
    //        return hWndParent;
    //}
}

BOOL CDialogBase::EnableDragAndDrop(BOOL bEnable)
{
    if (!this->hWndDialog)
    {
        this->bDragAcceptFileRequired = TRUE;
        return TRUE;
    }
    
    HMODULE hModule = ::GetModuleHandle(CDialogBase_SHELL_DLL);
    if (!hModule)
    {
        hModule =::LoadLibrary(CDialogBase_SHELL_DLL);
        if (!hModule)
            return FALSE;
    }

    typedef void (__stdcall *pfDragAcceptFiles)(HWND hWnd, BOOL fAccept);
    pfDragAcceptFiles pDragAcceptFiles;
    pDragAcceptFiles = (pfDragAcceptFiles) ::GetProcAddress(hModule,"DragAcceptFiles");
    if (!pDragAcceptFiles)
        return FALSE;
    pDragAcceptFiles(this->hWndDialog,bEnable); // DragAcceptFiles has no return value :(
    this->bDragAcceptFileStatusEnabled = bEnable;
    return TRUE;
}

CDialogBase::~CDialogBase(void)
{
    // if we are still in dialog box message loop, that means dialog is not closed
    // --> close it before destroying object to avoid invalid memory access and software crash
    if (this->bInsideDialogBoxFunction)
    {
        BOOL bCloseSuccess = this->CloseInternal(0,TRUE,TRUE);

        // wait end of Show method (aka DialogBoxParam function)
        while (this->bInsideDialogBoxFunction)
        {
            if (!bCloseSuccess)
                bCloseSuccess = this->CloseInternal(0,TRUE,TRUE);

            ::Sleep(500);
        }
    }
    ::EnterCriticalSection(&this->CriticalSection);
#ifdef _DEBUG
    if (this->CriticalSection.RecursionCount>1)
        BreakInDebugMode;
#endif

    ::DeleteCriticalSection(&this->CriticalSection);
}

//-----------------------------------------------------------------------------
// Name: GetComboboxEditHandle
// Object: get edit control associated to combobox
// Parameters :
//     in  : int nIDDlgItem : combobox id
//     out : 
//     return : edit handle
//-----------------------------------------------------------------------------
HWND CDialogBase::GetComboboxEditHandle(IN int nIDDlgItem)
{
    HWND hCombo= this->GetDlgItem(nIDDlgItem);
    COMBOBOXINFO info = {0};
    info.cbSize = sizeof(COMBOBOXINFO);
    ::GetComboBoxInfo(hCombo, &info);
    return info.hwndItem;
}

//-----------------------------------------------------------------------------
// Name: IsComboboxHandle
// Object: check if handle belongs to combobox (test combo, list and edit)
// Parameters :
//     in  : int nIDDlgItem : combobox id
//           HWND hItemToTest : handle to test
//     out : 
//     return : TRUE if handle belongs to combo
//-----------------------------------------------------------------------------
BOOL CDialogBase::IsComboboxHandle(IN int nIDDlgItem,IN HWND hItemToTest)
{
    HWND hCombo= this->GetDlgItem(nIDDlgItem);

    if (hCombo == hItemToTest)
        return TRUE;

    COMBOBOXINFO info = {0};
    info.cbSize = sizeof(COMBOBOXINFO);
    ::GetComboBoxInfo(hCombo, &info);
    if (info.hwndItem == hItemToTest)
        return TRUE;
    if (info.hwndList == hItemToTest)
        return TRUE;
    return FALSE;
}

BOOL CDialogBase::EnableRedraw(IN BOOL bEnable)
{
    return this->EnableRedrawFromHwnd(this->hWndDialog,bEnable);
}

BOOL CDialogBase::EnableRedraw(IN int nIDDlgItem,IN BOOL bEnable)
{
    HWND hWnd= this->GetDlgItem(nIDDlgItem);
    return this->EnableRedrawFromHwnd(hWnd,bEnable);
}

BOOL CDialogBase::EnableRedrawFromHwnd(IN HWND hWnd,IN BOOL bEnable)
{
    LRESULT RetValue;
    if (bEnable)
    {
        RetValue = ::SendMessage(hWnd, WM_SETREDRAW, TRUE, 0);
        // MSDN :
        // Note  RedrawWindow with the specified flags is used instead of InvalidateRect because the former is necessary for some controls 
        // that have nonclient area on their own, or have window styles that cause them to be given a nonclient area 
        // (such as WS_THICKFRAME, WS_BORDER, or WS_EX_CLIENTEDGE). If the control does not have a nonclient area, 
        // RedrawWindow with these flags will do only as much invalidation as InvalidateRect would.
        if (hWnd)
            ::RedrawWindow(hWnd, NULL, NULL, RDW_ERASE | RDW_FRAME | RDW_INVALIDATE | RDW_ALLCHILDREN);
    }
    else
    {
        // MSDN : 
        // If the application sends the WM_SETREDRAW message to a hidden window, the window becomes visible 
        // (that is, the operating system adds the WS_VISIBLE style to the window)
        RetValue = ::SendMessage(hWnd, WM_SETREDRAW, FALSE, 0);
    }
    // MSDN : An application returns zero if it processes this message
    return (RetValue == 0);
}

BOOL CDialogBase::Invalidate(IN int nIDDlgItem)
{
    return this->InvalidateFromHwnd(this->GetDlgItem(nIDDlgItem));
}
BOOL CDialogBase::InvalidateFromHwnd(HWND hItem)
{
    if (hItem)
        return ::InvalidateRect(hItem,NULL,FALSE);
    return FALSE;
}

BOOL CDialogBase::Redraw()
{
    return this->RedrawFromHwnd(this->hWndDialog);
}

BOOL CDialogBase::Redraw(IN int nIDDlgItem)
{
    return this->RedrawFromHwnd(this->GetDlgItem(nIDDlgItem));
}

//-----------------------------------------------------------------------------
// Name: RedrawFromHwnd
// Object: redraw an item
// Parameters :
//     in  : HWND hItem : handle of item to be redraw
//     out : 
//     return : 
//-----------------------------------------------------------------------------
BOOL CDialogBase::RedrawFromHwnd(HWND hItem)
{
    if (hItem)
        return ::RedrawWindow( hItem,NULL,NULL,RDW_INVALIDATE | RDW_UPDATENOW | RDW_ERASE | RDW_FRAME | RDW_ALLCHILDREN);
    return FALSE;
}

void CDialogBase::DialogUnitToPixels(RECT* pRect)
{
    ::MapDialogRect(this->hWndDialog,pRect);
}
void CDialogBase::PixelsToDialogUnit(RECT* pRect)
{
    /*
    MSDN:
    The MapDialogRect function assumes that the initial coordinates in the RECT structure represent dialog box units. To convert these coordinates from dialog box units to pixels, the function retrieves the current horizontal and vertical base units for the dialog box, then applies the following formulas: 
    left   = MulDiv(left,   baseunitX, 4);
    right  = MulDiv(right,  baseunitX, 4);
    top    = MulDiv(top,    baseunitY, 8);
    bottom = MulDiv(bottom, baseunitY, 8);

    If the dialog box template has the DS_SETFONT or DS_SHELLFONT style, the base units are the average width and height, in pixels, of the characters in the font specified by the template

    The GetDialogBaseUnits function retrieves the system's dialog base units, which are the average width and height of characters in the system font. For dialog boxes that use the system font, you can use these values to convert between dialog template units, as specified in dialog box templates, and pixels. For dialog boxes that do not use the system font, the conversion from dialog template units to pixels depends on the font used by the dialog box.
    For either type of dialog box, it is easier to use the MapDialogRect function to perform the conversion. MapDialogRect takes the font into account and correctly converts a rectangle from dialog template units into pixels

    The horizontal base unit returned by GetDialogBaseUnits is equal to the average width, in pixels, of the characters in the system font; the vertical base unit is equal to the height, in pixels, of the font. 
    For a dialog box that does not use the system font, the base units are the average width and height, in pixels, of the characters in the dialog's font. You can use the GetTextMetrics and GetTextExtentPoint32 functions to calculate these values for a selected font. However, by using the MapDialogRect function, you can avoid errors that might result if your calculations differ from those performed by the system.
    Each horizontal base unit is equal to 4 horizontal dialog template units; each vertical base unit is equal to 8 vertical dialog template units. Therefore, to convert dialog template units to pixels, use the following formulas: 
    pixelX = MulDiv(templateunitX, baseunitX, 4);
    pixelY = MulDiv(templateunitY, baseunitY, 8);

    Similarly, to convert from pixels to dialog template units, use the following formulas: 
    templateunitX = MulDiv(pixelX, 4, baseunitX);
    templateunitY = MulDiv(pixelY, 8, baseunitY);
    */

    LONG baseunitX;
    LONG baseunitY;
    
    // only for system dialog without font change
    //long DlgBaseUnit = GetDialogBaseUnits();
    //baseunitX= LOWORD(DlgBaseUnit);
    //baseunitY= HIWORD(DlgBaseUnit);

    // way to work for all
    RECT Rect={0,0,4,8};
    ::MapDialogRect(this->hWndDialog, &Rect);
    baseunitX = Rect.right;
    baseunitY = Rect.bottom;

    pRect->left   = ::MulDiv(pRect->left   , 4 ,baseunitX);
    pRect->right  = ::MulDiv(pRect->right  , 4 ,baseunitX);
    pRect->top    = ::MulDiv(pRect->top    , 8 ,baseunitY);
    pRect->bottom = ::MulDiv(pRect->bottom, 8 ,baseunitY);
}

//-----------------------------------------------------------------------------
// Name: ScreenToClient
// Object: translate lpRect from Screen coordinates to Client Coordinates
// Parameters :
//     in out : LPRECT lpRect : in  : Rect in screen coordinates
//                              out : Rect in dialog coordinates
//     return : 
//-----------------------------------------------------------------------------
BOOL CDialogBase::ScreenToClient(IN OUT LPRECT pRect)
{
    /*
    POINT p;
    p.x=lpRect->left;
    p.y=lpRect->top;
    ScreenToClient(hDlg,&p);
    lpRect->left=p.x;
    lpRect->top=p.y;
    p.x=lpRect->right;
    p.y=lpRect->bottom;
    ScreenToClient(hDlg,&p);
    lpRect->right=p.x;
    lpRect->bottom=p.y;
    */
    if (!::ScreenToClient (this->hWndDialog, (LPPOINT)pRect))
        return FALSE;
    if(!::ScreenToClient (this->hWndDialog, ((LPPOINT)pRect)+1))
        return FALSE;
    return TRUE;
}


//-----------------------------------------------------------------------------
// Name: GetClientWindowRect
// Object: get item rect in Client Coordinates
// Parameters :
//     in  : int nIDDlgItem : item id
//     out : LPRECT lpRect : Rect in dialog coordinates
//     return : 
//-----------------------------------------------------------------------------

BOOL CDialogBase::GetClientWindowRectFromHwnd(IN HWND hItem,OUT LPRECT lpRect)
{
    if (!::GetWindowRect(hItem,lpRect))
        return FALSE;
    if(!this->ScreenToClient(lpRect))
        return FALSE;
    return TRUE;
}

BOOL CDialogBase::GetClientWindowRect(IN int nIDDlgItem,OUT LPRECT lpRect)
{
    return this->GetClientWindowRectFromHwnd(this->GetDlgItem(nIDDlgItem),lpRect);
}

BOOL CDialogBase::OffsetPositionFromHwnd(HWND hWnd,LONG xOffset,LONG yOffset)
{
    RECT Rect;
    this->GetClientWindowRectFromHwnd(hWnd,&Rect);
    Rect.left+=xOffset;
    Rect.right+=xOffset;
    Rect.top+=yOffset;
    Rect.bottom+=yOffset;
    return this->SetRectFromHwnd(hWnd,&Rect);
}
BOOL CDialogBase::OffsetPositionFromId(int Id,LONG xOffset,LONG yOffset)
{
    return this->OffsetPositionFromHwnd(this->GetDlgItem(Id),xOffset,yOffset);
}

//-----------------------------------------------------------------------------
// Name: ParseGroupForAction
// Object: call specified callback for each item of the group
// Parameters :
//     in  : HWND hWndGroup : group window handle
//           CDialogBase::pfCallBackGroupAction CallBackFunction : callback called for each item
//           PVOID UserParam : parameter provided to callback
//     out : 
//     return : AND between all result of callback calls
//-----------------------------------------------------------------------------
BOOL CDialogBase::ParseGroupForAction(HWND hWndGroup,pfCallBackGroupAction CallBackFunction,PVOID UserParam)
{
    BOOL bRet=TRUE;
    BOOL bFunctionRet=TRUE;
    if (!hWndGroup)
        return FALSE;
    HWND hWndCtrl;
    hWndCtrl=hWndGroup;
    // for each item of the group
    for(;;)
    {
        // call callback with item handle and UserParam
        bFunctionRet = CallBackFunction(hWndCtrl,UserParam);
        bRet=bRet && bFunctionRet;

        // get next window handle
        hWndCtrl = GetWindow( hWndCtrl, GW_HWNDNEXT );
        // if no next window or next window has WS_GROUP style (window outside of the current group)
        if ( (hWndCtrl==NULL) || (GetWindowLong(hWndCtrl,GWL_STYLE)&WS_GROUP))
            break;

    } 
    return bRet;
}

//-----------------------------------------------------------------------------
// Name: CallBackRedrawGroup
// Object: called to redraw item of a group. Call CDialogBase::RedrawFromHwnd
// Parameters :
//     in  : HWND Item : window handle
//     out : 
//     return : TRUE
//-----------------------------------------------------------------------------
BOOL CDialogBase::CallBackRedrawGroup(IN HWND Item,IN PVOID UserParam)
{
    CDialogBase* pDialogBase = (CDialogBase*)UserParam;
    pDialogBase->RedrawFromHwnd(Item);
    return TRUE;
}

//-----------------------------------------------------------------------------
// Name: RedrawGroupFromHwnd
// Object: Call CDialogBase::RedrawFromHwnd for each item of a group
// Parameters :
//     in  : HWND hWndGroup : group window handle
//     out : 
//     return : TRUE
//-----------------------------------------------------------------------------
BOOL CDialogBase::RedrawGroup(IN int nIDDlgItem)
{
    return this->RedrawGroupFromHwnd(this->GetDlgItem(nIDDlgItem));
}

BOOL CDialogBase::RedrawGroupFromHwnd(IN HWND hWndGroup)
{
    if(!hWndGroup)
        return FALSE;
    return this->ParseGroupForAction(hWndGroup,CDialogBase::CallBackRedrawGroup,this);
}

//-----------------------------------------------------------------------------
// Name: CallBackShowGroup
// Object: called to show item of a group. Call ShowWindow API
// Parameters :
//     in  : HWND Item : window handle
//           PVOID UserParam : int nCmdShow of ShowWindow API
//     out : 
//     return : TRUE
//-----------------------------------------------------------------------------
BOOL CDialogBase::CallBackShowGroup(HWND Item,PVOID UserParam)
{
    ::ShowWindow(Item,(int)UserParam);
    return TRUE;
}

//-----------------------------------------------------------------------------
// Name: ShowGroup
// Object: Call ShowWindow API for each item of a group
// Parameters :
//     in  : HWND hWndGroup : group window handle
//           int CmdShow : int nCmdShow parameter of ShowWindow API
//     out : 
//     return : TRUE
//-----------------------------------------------------------------------------
BOOL CDialogBase::ShowGroup(IN int nIDDlgItem,IN int CmdShow)
{
    HWND hWndGroup = this->GetDlgItem(nIDDlgItem);
    return this->ShowGroupFromHandle(hWndGroup,CmdShow);
}

BOOL CDialogBase::ShowGroupFromHandle(IN HWND hWndGroup,IN int CmdShow)
{
    if (!hWndGroup)
        return FALSE;
    return CDialogBase::ParseGroupForAction(hWndGroup,CDialogBase::CallBackShowGroup,(PVOID)CmdShow);
}

//-----------------------------------------------------------------------------
// Name: CallBackShowGroup
// Object: called to enable item of a group. Call ShowWindow API
// Parameters :
//     in  : HWND Item : window handle
//           PVOID UserParam : BOOL bEnable parameter of EnableWindow API
//     out : 
//     return : TRUE
//-----------------------------------------------------------------------------
BOOL CDialogBase::CallBackEnableGroup(HWND Item,PVOID UserParam)
{
    ::EnableWindow(Item,(BOOL)UserParam);
    return TRUE;
}

//-----------------------------------------------------------------------------
// Name: EnableGroup
// Object: Call EnableWindow API for each item of a group
// Parameters :
//     in  : HWND hWndGroup : group window handle
//           BOOL bEnable : BOOL bEnable parameter of EnableWindow API
//     out : 
//     return : TRUE
//-----------------------------------------------------------------------------
BOOL CDialogBase::EnableGroup(IN int nIDDlgItem,IN BOOL bEnable)
{
    HWND hWndGroup = this->GetDlgItem(nIDDlgItem);
    return this->EnableGroupFromHandle(hWndGroup,bEnable);
}
BOOL CDialogBase::EnableGroupFromHandle(IN HWND hWndGroup,IN BOOL bEnable)
{
    if (!hWndGroup)
        return FALSE;
    return CDialogBase::ParseGroupForAction(hWndGroup,CDialogBase::CallBackEnableGroup,(PVOID)bEnable);
}

//-----------------------------------------------------------------------------
// Name: CallBackMoveGroup
// Object: called to move items of a group. Call SetWindowPos API
// Parameters :
//     in  : HWND Item : window handle
//           PVOID UserParam : pointer to a MOVE_GROUP_STRUCT containing 
//                              informations to set window position
//     out : 
//     return : TRUE on success
//-----------------------------------------------------------------------------
BOOL CDialogBase::CallBackMoveGroup(HWND Item,PVOID UserParam)
{
    CDialogBase::MOVE_GROUP_STRUCT* pMoveStruct;
    RECT Rect;
    // get struct pointer
    pMoveStruct=(MOVE_GROUP_STRUCT*)UserParam;

    // get old item position
    if(!pMoveStruct->pDialogBase->GetClientWindowRectFromHwnd(Item,&Rect))
        return FALSE;
    // set new position applying displacement
    return ::SetWindowPos(Item,NULL,
                          Rect.left+pMoveStruct->Point.x,Rect.top+pMoveStruct->Point.y,
                          0,0,
                          SWP_NOACTIVATE|SWP_NOZORDER|SWP_NOSIZE);
}

//-----------------------------------------------------------------------------
// Name: CallBackMoveGroup
// Object: called to move items of a group. Call SetWindowPos API
// Parameters :
//     in  : HWND hClient : handle client window from which point refer (generally handle to dialog)
//           HWND hWndGroup : handle to group window
//           POINT* pPoint : new group position in client coordinates
//     out : 
//     return : TRUE on success
//-----------------------------------------------------------------------------
BOOL CDialogBase::MoveGroupTo(IN int nIDDlgItem,IN POINT* pPoint)
{
    HWND hWndGroup = this->GetDlgItem(nIDDlgItem);
    return this->MoveGroupToFromHandle(hWndGroup,pPoint);
}
BOOL CDialogBase::MoveGroupToFromHandle(IN HWND hWndGroup,IN POINT* pPoint)
{
    RECT Rect;
    MOVE_GROUP_STRUCT MoveStruct;
    BOOL bRet;


    if (!hWndGroup)
        return FALSE;

    // get window group Rect
    this->GetClientWindowRectFromHwnd(hWndGroup,&Rect);

    // fill MOVE_GROUP_STRUCT
    MoveStruct.pDialogBase=this;
    // compute displacement values
    MoveStruct.Point.x=pPoint->x-Rect.left;
    MoveStruct.Point.y=pPoint->y-Rect.top;

    // apply displacement for all items of the group
    bRet=CDialogBase::ParseGroupForAction(hWndGroup,CDialogBase::CallBackMoveGroup,&MoveStruct);

    // redraw group
    this->RedrawFromHwnd(hWndGroup);

    return bRet;
}

BOOL CDialogBase::OffsetGroupPosition(IN int nIDDlgItem,IN LONG xOffset,IN LONG yOffset)
{
    return this->OffsetGroupPositionFromHandle(this->GetDlgItem(nIDDlgItem),xOffset,yOffset);
}

BOOL CDialogBase::OffsetGroupPositionFromHandle(IN HWND hWndGroup,IN LONG xOffset,IN LONG yOffset)
{
    RECT Rect;
    this->GetClientWindowRectFromHwnd(hWndGroup,&Rect);
    POINT pt;
    pt.x = Rect.left+xOffset;
    pt.y = Rect.top+yOffset;
    return this->MoveGroupToFromHandle(hWndGroup,&pt);
}

HICON CDialogBase::SetStaticIcon(int StaticControlId,int IconId,int Width,int Height)
{
    HICON hIcon = this->LoadIcon(IconId,Width,Height);
    if (!hIcon)
        return hIcon;
    this->SetStaticIcon(StaticControlId,hIcon);
    return hIcon;
}

HICON CDialogBase::SetDlgButtonIconFromHwnd(HWND hWnd,int IconId,int Width,int Height)
{
    HICON hIcon = this->LoadIcon(IconId,Width,Height);
    if (!hIcon)
        return hIcon;
    this->SetDlgButtonIconFromHwnd(hWnd,hIcon);
    return hIcon;
}



BOOL CDialogBase::WaitDialogEnd(DWORD TimeOutInMs)
{
    while (this->bInsideDialogBoxFunction)
    {
        ::Sleep(100);
        if (TimeOutInMs!=INFINITE)
        {
            if (TimeOutInMs>100)
                TimeOutInMs-=100;
            else
                FALSE;
        }
    }
    return TRUE;
}

BOOL CDialogBase::WaitDialogThreadEnd(DWORD TimeOutInMs)
{
    // Notice : assume to take only static var, so dialog object can be destroyed
    BOOL Ret;
    // get thread owing dialog proc message loop
    DWORD WindowThreadId = ::GetWindowThreadProcessId(this->hWndDialog,NULL);
    if (WindowThreadId==0)// GetWindowThreadProcessId return 0 if hWnd is invalid
        return FALSE;

    // avoid to wait for our self
    if (::GetCurrentThreadId() == WindowThreadId)
        return FALSE;

    // open thread for synchronization
    HANDLE hThread = ::OpenThread(SYNCHRONIZE,FALSE,WindowThreadId);
    if (!hThread)
        return FALSE;

    // wait until thread end or time out
    Ret = (::WaitForSingleObject(hThread,TimeOutInMs)== WAIT_OBJECT_0);

    ::CloseHandle(hThread);

    return Ret;
}


void CDialogBase::Flash()
{
    FLASHWINFO FlashInfo={0};
    FlashInfo.cbSize = sizeof(FLASHWINFO);
    FlashInfo.hwnd = this->hWndDialog;
    FlashInfo.dwFlags = FLASHW_ALL;
    FlashInfo.uCount = 3;
    // MSDN : dwTimeout Rate at which the window is to be flashed, in milliseconds. 
    //                  If dwTimeout is zero, the function uses the default cursor blink rate
    FlashInfo.dwTimeout = 0;

    ::FlashWindowEx(&FlashInfo);
}


//-----------------------------------------------------------------------------
// Name: ReportMessage
// Object: display messagebox
// Parameters :
//     in  :
//           TCHAR* pszMsg : message to display
//           ReportMessageTypes MsgType : type of message (information, error , warning)
//     return : MessageBox return value
//-----------------------------------------------------------------------------
int CDialogBase::ReportMessage(TCHAR* pszMsg, ReportMessageTypes MsgType)
{
    return this->ReportMessageEx(pszMsg, NULL, MsgType);
}

void CDialogBase::GetDefaultCaptionAndFlagFromReportMessageType(IN ReportMessageTypes MsgType,
                                                                OUT const TCHAR** ppszCaption,
                                                                OUT UINT* pMessageBoxFlag
                                                                )
{
    TCHAR* pszCaption;
    UINT MessageBoxFlag = 0;
    switch(MsgType)
    {
    default:
    case ReportMessageType_ERROR:
        pszCaption = _T("Error");
        MessageBoxFlag = MB_OK|MB_ICONERROR;
        break;
    case ReportMessageType_ERROR_QUESTION_YES_NO:
        pszCaption = _T("Error");
        MessageBoxFlag = MB_ICONERROR | MB_YESNO;
        break;
    case ReportMessageType_WARNING:
        pszCaption = _T("Warning");
        MessageBoxFlag = MB_OK|MB_ICONWARNING;
        break;
    case ReportMessageType_WARNING_QUESTION_YES_NO:
        pszCaption = _T("Warning");
        MessageBoxFlag = MB_ICONWARNING | MB_YESNO;
        break;
    case ReportMessageType_INFO:
        pszCaption = _T("Information");
        MessageBoxFlag = MB_OK|MB_ICONINFORMATION;
        break;
    case ReportMessageType_INFO_QUESTION_YES_NO:
        pszCaption = _T("Information");
        MessageBoxFlag = MB_ICONINFORMATION | MB_YESNO;
        break;
    case ReportMessageType_QUESTION_YES_NO:
        pszCaption = _T("Question");
        MessageBoxFlag = MB_ICONQUESTION | MB_YESNO;
        break;
    case ReportMessageType_QUESTION_YES_NO_CANCEL:
        pszCaption = _T("Question");
        MessageBoxFlag = MB_ICONQUESTION | MB_YESNOCANCEL;
        break;
    }

    *ppszCaption = pszCaption;
    *pMessageBoxFlag = MessageBoxFlag;
}

int CDialogBase::ReportMessageEx(TCHAR* pszMsg, TCHAR* OptCaption, ReportMessageTypes MsgType)
{
    TCHAR* pszCaption;
    UINT MessageBoxFlag = 0;

    this->GetDefaultCaptionAndFlagFromReportMessageType(MsgType,(const TCHAR**)&pszCaption,&MessageBoxFlag);

    if (OptCaption)
        pszCaption = OptCaption;

    if (!this->bModalDialog)
        this->EnableClosingCount(FALSE);
    int iRet = ::MessageBox(this->hWndDialog,pszMsg,pszCaption,MessageBoxFlag);
    if (!this->bModalDialog)
        this->EnableClosingCount(TRUE);
    return iRet;
}

//-----------------------------------------------------------------------------
// Name: ReportError
// Object: display error message
// Parameters :
//     in  : 
//           TCHAR* pszMsg : error message to display
//     return : MessageBox return value
//-----------------------------------------------------------------------------
int CDialogBase::ReportError(TCHAR* pszMsg)
{
    return this->ReportMessage(pszMsg,ReportMessageType_ERROR);
}

//-----------------------------------------------------------------------------
// Name: Show
// Object: show the dialog box
// Parameters :
//     in  : HINSTANCE hInstance : application instance
//           HWND hWndDialog : parent window dialog handle
//           int ResourceId : dialog resource id 
//           int IconResourceId : dialog icon resource id
//     out :
//     return : DialogBoxParam return value
//-----------------------------------------------------------------------------
INT_PTR CDialogBase::Show(HINSTANCE hInstance,HWND hWndParentDialog,int ResourceId,int IconResourceId)
{
    this->IconResourceId = IconResourceId;
    this->hInstance = hInstance;
    this->hWndParentDialog = hWndParentDialog;
    this->bInsideDialogBoxFunction = TRUE;
    this->bModalDialog = TRUE;
    INT_PTR DialogResult = ::DialogBoxParam(hInstance,MAKEINTRESOURCE(ResourceId),hWndParentDialog,CDialogBase::DlgProc,(LPARAM)this);
    // must be last object access (see destructor code)
    this->bInsideDialogBoxFunction = FALSE;

    return DialogResult;
}

//-----------------------------------------------------------------------------
// Name: ShowIndirect
// Object: show the dialog box from resource template
// Parameters :
//     in  : HINSTANCE hInstance : application instance
//           HWND hWndDialog : parent window dialog handle
//           LPCDLGTEMPLATE hDialogTemplate : buffer containing dialog resource template
//           int IconResourceId : dialog icon resource id
//     out :
//     return : DialogBoxParam return value
//-----------------------------------------------------------------------------
INT_PTR CDialogBase::ShowIndirect(HINSTANCE hInstance,HWND hWndParentDialog,LPCDLGTEMPLATE hDialogTemplate,int IconResourceId)
{
    this->IconResourceId = IconResourceId;
    this->hInstance = hInstance;
    this->hWndParentDialog = hWndParentDialog;
    this->bInsideDialogBoxFunction = TRUE;
    this->bModalDialog = TRUE;
    INT_PTR DialogResult = ::DialogBoxIndirectParam(hInstance,hDialogTemplate,hWndParentDialog,CDialogBase::DlgProc,(LPARAM)this);
    // must be last object access (see destructor code)
    this->bInsideDialogBoxFunction = FALSE;

    return DialogResult;
}

//-----------------------------------------------------------------------------
// Name: Create modless dialog (initially hidden : depends of WS_VISIBLE)
// Object: Create the dialog box
// Parameters :
//     in  : HINSTANCE hInstance : application instance
//           HWND hWndDialog : main window dialog handle
//           int ResourceId : dialog resource id 
//           int IconResourceId : dialog icon resource id
//     out :
//     return : CreateDialogParam return value
//-----------------------------------------------------------------------------
HWND CDialogBase::CreateModless(HINSTANCE hInstance,HWND hWndParentDialog,int ResourceId,int IconResourceId)
{
    this->IconResourceId = IconResourceId;
    this->hInstance = hInstance;
    this->hWndParentDialog = hWndParentDialog;
    this->bModalDialog = FALSE;
    this->bInsideDialogBoxFunction = TRUE;
    // this->bCloseModalSubDialogsIfAny = TRUE; doesn't solve anything
    HWND hWndCreated = ::CreateDialogParam(hInstance,MAKEINTRESOURCE(ResourceId),hWndParentDialog,CDialogBase::DlgProc,(LPARAM)this);
    if (hWndCreated == NULL)
        this->bInsideDialogBoxFunction = FALSE;
    return hWndCreated;
}

//-----------------------------------------------------------------------------
// Name: Create modless dialog indirect (initially hidden : depends of WS_VISIBLE)
// Object: Create the dialog box
// Parameters :
//     in  : HINSTANCE hInstance : application instance
//           HWND hWndDialog : main window dialog handle
//           LPCDLGTEMPLATE hDialogTemplate : buffer containing dialog resource template
//           int IconResourceId : dialog icon resource id
//     out :
//     return : CreateDialogIndirectParam return value
//-----------------------------------------------------------------------------
HWND CDialogBase::CreateModlessIndirect(HINSTANCE hInstance,HWND hWndParentDialog,LPCDLGTEMPLATE hDialogTemplate,int IconResourceId)
{
    this->IconResourceId = IconResourceId;
    this->hInstance = hInstance;
    this->hWndParentDialog = hWndParentDialog;
    this->bModalDialog = FALSE;
    this->bInsideDialogBoxFunction = TRUE;
    // this->bCloseModalSubDialogsIfAny = TRUE; doesn't solve anything
    HWND hWndCreated = ::CreateDialogIndirectParam(hInstance,hDialogTemplate,hWndParentDialog,CDialogBase::DlgProc,(LPARAM)this);
    if (hWndCreated == NULL)
        this->bInsideDialogBoxFunction = FALSE;
    return hWndCreated;
}

//-----------------------------------------------------------------------------
// Name: Close
// Object: close the dialog box
// Parameters :
//     in  : int Code : dialog box return value
//     out :
//     return : TRUE on success, FALSE on error
//-----------------------------------------------------------------------------
BOOL CDialogBase::Close(INT_PTR Code)
{
    return this->CloseInternal(Code,FALSE,FALSE);
}

BOOL CDialogBase::CloseEx(INT_PTR Code,BOOL bCloseModalSubDialogsIfAny)
{
    return this->CloseInternal(Code,bCloseModalSubDialogsIfAny,FALSE);
}

BOOL CDialogBase::CloseInternal(INT_PTR Code,BOOL bCloseModalSubDialogsIfAny, BOOL bForceClosing)
{
    // if close has already been called
    if (this->hWndDialog == NULL)
        return TRUE;

    if (this->DisableClosingCount>0)
    {
        this->Flash();
        BreakInDebugMode;
        return FALSE;
    }

    if (bCloseModalSubDialogsIfAny)
        this->CloseModalSubDialogsIfAny();

    if(!this->OnCloseInternal())
    {
        if (!bForceClosing)
            return FALSE;
    }

    if (this->bDragAcceptFileStatusEnabled)
        this->EnableDragAndDrop(FALSE);

    ::SetWindowLongPtr(this->hWndDialog,GWLP_USERDATA,(LONG_PTR)NULL);

#ifdef _DEBUG
    DWORD dwLastError = 0;
#endif
    BOOL bRet = TRUE;

    if (this->IsModal())
    {
        bRet = ::EndDialog(this->hWndDialog,Code);
#ifdef _DEBUG
        if (!bRet)
        {
            dwLastError = ::GetLastError();
            BreakInDebugMode;
        }
#endif
    }
    else
    {
        // modeless dialog --> no return --> "Code" parameter is useless
        bRet = ::DestroyWindow(this->hWndDialog);

        // must be last object access (see destructor code)
        this->bInsideDialogBoxFunction = FALSE;

#ifdef _DEBUG
        if (!bRet)
        {
            dwLastError = ::GetLastError();
            // common errors : 
            //  5 : close must be done in same thread that window creation
            BreakInDebugMode;
        }
#endif
    }
    if (bRet)
        this->hWndDialog = NULL;
    return bRet;
}


typedef struct  
{
    CDialogBase* pDialogBase;
    HWND hWndPopWindow;
}EnumWndProc_CloseModalSubDialogsIfAny_PARAM;

BOOL CALLBACK CDialogBase::EnumWndProc_CloseModalSubDialogsIfAny(HWND hWnd,LPARAM lParam)
{
    // CDialogBase* pDialogBase = (CDialogBase*)lParam;
    EnumWndProc_CloseModalSubDialogsIfAny_PARAM* pParam = (EnumWndProc_CloseModalSubDialogsIfAny_PARAM*)lParam;
    CDialogBase* pDialogBase = pParam->pDialogBase;
    // HWND hWndSearchedParent = pParam->hWndPopWindow;
    // check if window is a dialog
    TCHAR ClassName[64];
    HWND hWndCurrentDialog = pDialogBase->GetControlHandle();
    ::RealGetWindowClass(hWnd,ClassName,_countof(ClassName));
    if ( _tcscmp(ClassName,TEXT("#32770")) == 0)
    {
        // if it is our own window
        if (hWndCurrentDialog == hWnd)
            return TRUE;

        // close only popup dialogs, not child dialog (it will by done by the owner)
        if ( ::GetWindowLongPtr(hWnd ,GWL_STYLE) & WS_POPUP )
        {
            // parent checking can't be used in case NULL parent was specified
            //HWND hParent = hWnd;
            //while (hParent)
            //{
            //    hParent = ::GetParent(hParent);
            //    if (hParent == hWndSearchedParent)
            //    {
                    //TCHAR StrMessage[512];
                    //_sntprintf(StrMessage,_countof(StrMessage),TEXT("Dialog:0x%p\r\n"),hWnd);
                    //::OutputDebugString(StrMessage);
                    //::PostMessage(hWnd,WM_CLOSE,0,0);
                    ::SendMessage(hWnd,WM_CLOSE,0,0); // do not use post, force sub dialog closing before closing ourself
                                                      // quite useless anyway as our close will occur before the end of the object method that has created the popup
                                                      // so if object is deleted, the method object will crash after the return of the Dialog
                                                      // we have just order the OnClose events
                                                      // A Modeless A->ShowModalB creates B modal
                                                      // Parent of A Closing --> A->Close --> B->Close --> B->OnClose
                                                      // A->OnClose
                                                      // End of A->ShowModalB --> Crash
                    return TRUE;
            //    }
            //}

        }
    }
    return TRUE;
}

BOOL CDialogBase::CloseModalSubDialogsIfAny()
{
    EnumWndProc_CloseModalSubDialogsIfAny_PARAM Param;
    Param.pDialogBase = this;
    // if dialog isn't a popup dialog and a dialog has been created, created dialog parent will be the popup parent dialog of current dialog
    Param.hWndPopWindow = (::GetWindowLongPtr(this->hWndDialog ,GWL_STYLE) & WS_POPUP ) ? this->hWndDialog : this->GetParentPopUpDialog();
    // modal sub windows are in same thread (else current dialog message loop is not blocked by other window, so there is no interest to close them)

    // EnumThreadWindows : Enumerates all nonchild windows associated with a thread by passing the handle to each window, in turn, to an application-defined callback function.
    // EnumThreadWindows : It returns them in Z order. First the top-most window with WS_EX_TOPMOST set, until the bottom-most window with WS_EX_TOPMOST set, then the top-most window without WS_EX_TOPMOST, though to the bottom-most window without WS_EX_TOPMOST. Note that visibility is not a determining factor, so an invisible window that's higher in the Z-order than a visible window will still appear before it.
    return ::EnumThreadWindows(this->GetWindowThreadID(),CDialogBase::EnumWndProc_CloseModalSubDialogsIfAny,(LPARAM)&Param);
}

BOOL CDialogBase::CloseFromOtherThread(BOOL bCloseModalSubDialogsIfAny,BOOL bWaitEndOfDialog,DWORD WaitTimeout)
{
    if (this->DisableClosingCount>0)
    {
        this->Flash();
        BreakInDebugMode;
        return FALSE;
    }

    // avoid destructor to be called when this function is not completed (no AddRef / Release currently)
    ::EnterCriticalSection(&this->CriticalSection);

    BOOL bRetValue = TRUE;
    if (bCloseModalSubDialogsIfAny)
        this->CloseModalSubDialogsIfAny();

    ::PostMessage(this->hWndDialog,WM_CLOSE,0,0);

    if (bWaitEndOfDialog)
        bRetValue = this->WaitDialogEnd(WaitTimeout);

    ::LeaveCriticalSection(&this->CriticalSection);

    return bRetValue;
}

typedef struct 
{
    CDialogBase* pDialogBase;
    HWND hWndPopWindow;
    HWND hWndFoundPopWindow;
}EnumWndProc_GetTopMostModalPopUpSubDialogs_PARAM;

BOOL CALLBACK CDialogBase::EnumWndProc_GetTopMostModalPopUpSubDialogs(HWND hWnd,LPARAM lParam)
{
    // CDialogBase* pDialogBase = (CDialogBase*)lParam;
    EnumWndProc_GetTopMostModalPopUpSubDialogs_PARAM* pParam = (EnumWndProc_GetTopMostModalPopUpSubDialogs_PARAM*)lParam;
    CDialogBase* pDialogBase = pParam->pDialogBase;
    HWND hWndCurrentDialog = pDialogBase->GetControlHandle();
    // HWND hWndSearchedParent = pParam->hWndPopWindow;
    // check if window is a dialog
    TCHAR ClassName[64];
 
    if (hWnd == hWndCurrentDialog)
        return TRUE;
 
    ::RealGetWindowClass(hWnd,ClassName,_countof(ClassName));
    if ( _tcscmp(ClassName,TEXT("#32770")) == 0)
    {
        if ( ::GetWindowLongPtr(hWnd ,GWL_STYLE) & WS_POPUP )
        {
            // parent checking can't be used in case NULL parent was specified
            //HWND hParent = hWnd;
            //while (hParent)
            //{
            //    hParent = ::GetParent(hParent);
            //    if (hParent == hWndSearchedParent)
            //    {
                    pParam->hWndFoundPopWindow = hWnd;
                    return FALSE;
            //    }
            //}
        }
    }
    return TRUE;
}

HWND CDialogBase::GetTopMostModalPopUpSubDialogs()
{
    EnumWndProc_GetTopMostModalPopUpSubDialogs_PARAM Param;
    //    Param.pDialogBase = this;
    // if dialog isn't a popup dialog and a dialog has been created, created dialog parent will be the popup parent dialog of current dialog
    Param.hWndPopWindow = (::GetWindowLongPtr(this->hWndDialog ,GWL_STYLE) & WS_POPUP ) ? this->hWndDialog : this->GetParentPopUpDialog();
    Param.hWndFoundPopWindow = 0;
    Param.pDialogBase = this;
    // modal sub windows are in same thread (else current dialog message loop is not blocked by other window, so there is no interest to close them)
 
    // EnumThreadWindows : Enumerates all nonchild windows associated with a thread by passing the handle to each window, in turn, to an application-defined callback function.
    // EnumThreadWindows : It returns them in Z order. First the top-most window with WS_EX_TOPMOST set, until the bottom-most window with WS_EX_TOPMOST set, then the top-most window without WS_EX_TOPMOST, though to the bottom-most window without WS_EX_TOPMOST. Note that visibility is not a determining factor, so an invisible window that's higher in the Z-order than a visible window will still appear before it.
    ::EnumThreadWindows(this->GetWindowThreadID(),CDialogBase::EnumWndProc_GetTopMostModalPopUpSubDialogs,(LPARAM)&Param);
 
    return Param.hWndFoundPopWindow;
}

void CDialogBase::SetWaitCursor(BOOL bWait)
{
    if (bWait)
    {
        if (!this->hWaitCursor)
            this->hWaitCursor=::LoadCursor(NULL,IDC_WAIT);
    }
    this->bWaitCursor = bWait;

    // force cursor update
    POINT pt;
    GetCursorPos(&pt);
    SetCursorPos(pt.x,pt.y);
}

void CDialogBase::SetDlgIcon(HINSTANCE hInstance,int IconResourceId)
{
    ::SendMessage(this->hWndDialog, WM_SETICON, ICON_BIG,
                    (LPARAM)::LoadImage(hInstance, MAKEINTRESOURCE(IconResourceId),
                                        IMAGE_ICON,::GetSystemMetrics(SM_CXICON),::GetSystemMetrics(SM_CYICON),
                                        LR_DEFAULTCOLOR|LR_SHARED)
                    );
    ::SendMessage(this->hWndDialog, WM_SETICON, ICON_SMALL,
                    (LPARAM)::LoadImage(hInstance, MAKEINTRESOURCE(IconResourceId),
                                        IMAGE_ICON,::GetSystemMetrics(SM_CXSMICON),::GetSystemMetrics(SM_CYSMICON),
                                        LR_DEFAULTCOLOR|LR_SHARED)
                    );
}

DWORD CDialogBase::GetHeight()
{
    RECT Rect;
    if (::GetWindowRect(this->hWndDialog,&Rect))
    {
        return Rect.bottom-Rect.top;
    }
    //else
    return 0;
}
DWORD CDialogBase::GetWidth()
{
    RECT Rect;
    if (::GetWindowRect(this->hWndDialog,&Rect))
    {
        return Rect.right-Rect.left;
    }
    //else
    return 0;
}
DWORD CDialogBase::GetInternalHeight()
{
    RECT Rect;
    if (this->GetInternalRect(&Rect))
    {
        return Rect.bottom-Rect.top;
    }
    //else
    return 0;
}
DWORD CDialogBase::GetInternalWidth()
{
    RECT Rect;
    if (this->GetInternalRect(&Rect))
    {
        return Rect.right-Rect.left;
    }
    //else
    return 0;
}

void CDialogBase::CenterOwner()
{
    this->CenterOnWindow(this->hWndParentDialog);
}

void CDialogBase::CenterOnWindow(HWND hWnd)
{
    // center dialog
    RECT Rect;
    RECT ParentRect;
    ::GetWindowRect(this->hWndDialog,&Rect);

    ::GetWindowRect(hWnd,&ParentRect);
    Rect.left = ParentRect.left+(ParentRect.right-ParentRect.left-(Rect.right-Rect.left) )/2; // Warning : force int cast to make sign div
    Rect.top = ParentRect.top+(ParentRect.bottom-ParentRect.top-(Rect.bottom-Rect.top) )/2; // Warning : force int cast to make sign div

    ::SetWindowPos(this->hWndDialog, 0, Rect.left,Rect.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

//-----------------------------------------------------------------------------
// Name: GetAssociatedObject
// Object: Get object associated to window handle
// Parameters :
//     in  : HWND hWndDialog : handle of the window
//     out :
//     return : associated object if found, NULL if not found
//-----------------------------------------------------------------------------
CDialogBase* CDialogBase::GetAssociatedObject(HWND hWndDialog)
{
    return (CDialogBase*)GetWindowLongPtr(hWndDialog,GWLP_USERDATA);
}
//-----------------------------------------------------------------------------
// Name: DlgProc
// Object: dialog callback of the dump dialog
// Parameters :
//     in  : 
//     out :
//     return : 
//-----------------------------------------------------------------------------
INT_PTR CALLBACK CDialogBase::DlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
    case WM_INITDIALOG:
        {
            CDialogBase* pDialogBase=(CDialogBase*)lParam;
            pDialogBase->hWndDialog=hWnd;

            ::SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)pDialogBase);

            if (pDialogBase->bDragAcceptFileRequired)
                pDialogBase->EnableDragAndDrop(TRUE);

            // load dlg icons
            pDialogBase->SetDlgIcon(pDialogBase->hInstance,pDialogBase->IconResourceId);

            BOOL bRet = pDialogBase->OnInitInternal();
            pDialogBase->bInitDone=TRUE;

            return bRet;
        }
        break;
    case WM_CLOSE:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            // 1) specific case as EDIT control in multi lines configuration send WM_CLOSE directly when edit has focus and VK_ESCAPE is pressed (known windows bug for backward compatibility)
            // 
            //LRESULT CDialogBaseChild::OnMessage(UINT uMsg,WPARAM wParam,LPARAM lParam)
            //{
            //    // specific case as EDIT control in multi lines configuration send WM_CLOSE directly when edit has focus and VK_ESCAPE is pressed (known windows bug for backward compatibility)
            //    switch (uMsg)
            //    {
            //    case WM_CLOSE:
            //        return TRUE;
            //    case WM_SYSCOMMAND:
            //        {
            //            if ( wParam == SC_CLOSE) 
            //            {
            //                this->Close();
            //            }
            //        }
            //    }
            //    return FALSE;
            //}

            // 2) take care of buttons with IDCANCEL because they send WM_CLOSE too
            if (pDialogBase->OnMessage(uMsg,wParam,lParam)) // if WM_CLOSE returns TRUE, we should not treat it
                return TRUE;
            pDialogBase->Close();
        }
        break;

    case WM_COMMAND:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnCommand(wParam,lParam);
        }
        break;

    case WM_NOTIFY:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                return FALSE;
            return pDialogBase->OnNotify(wParam,lParam);
        }
        break;
    case WM_DROPFILES:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnDropFiles(wParam,lParam);
        }
        break;
    case WM_LBUTTONDOWN:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnMouseDown(wParam,lParam);
        }
        break;

    case WM_LBUTTONUP:
    case WM_KILLFOCUS:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnMouseUp(wParam,lParam);
        }
        break;

    case WM_MOUSEMOVE:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnMouseMove(wParam,lParam);
        }
        break;
    case WM_SIZING:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnSizing(wParam,lParam);
        }
        break;
    case WM_SIZE:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnSize(wParam,lParam);
        }
        break;
    case WM_SETCURSOR:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;

            if (pDialogBase->bWaitCursor)
            {
                ::SetCursor(pDialogBase->hWaitCursor);
                //must use SetWindowLongPtr to return value from dialog proc
                ::SetWindowLongPtr(hWnd, DWLP_MSGRESULT, TRUE);
                return TRUE;
            }
            else
            {
                ::SetWindowLongPtr(hWnd, DWLP_MSGRESULT, FALSE);
                return FALSE;
            }
            break;
        }
    case WM_ACTIVATE:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->OnActivate(wParam,lParam);
        }
        break;
    case WM_DIALOG_BASE_ENABLE_CLOSING:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                break;
            pDialogBase->EnableClosingCount((BOOL) wParam);
        }
        break;
    default:
        {
            CDialogBase* pDialogBase=CDialogBase::GetAssociatedObject(hWnd);
            if (!pDialogBase)
                return FALSE;// we must return FALSE to tell that message is not handled
            // Notice: the following will this will fail if message should return FALSE
            LRESULT WndRet = pDialogBase->OnMessage(uMsg,wParam,lParam);
            if (WndRet)
            {
                //Typically, the dialog box procedure should return TRUE if it processed the message, and FALSE if it did not. 
                // If the dialog box procedure returns FALSE, the dialog manager performs the default dialog operation in response to the message.
                // If the dialog box procedure processes a message that requires a specific return value, the dialog box procedure should set the desired 
                // return value by calling 
                ::SetWindowLongPtr(pDialogBase->hWndDialog,DWLP_MSGRESULT,WndRet);
            }
            return WndRet;
        }
        // return FALSE;
    }
    return TRUE;
}

//-----------------------------------------------------------------------------
// Name: EnableClosingCount 
// Object: enable or disable closing. 
//          the number of disabled closing is counted, that means
//          each EnableClosingCount(FALSE) must be balanced by a EnableClosingCount(TRUE)
//      usage : 
//          EnableClosingCount(FALSE)
//          dialog closing is disabled
//          EnableClosingCount(TRUE)
//          the dialog can be closed again
// Parameters :
//     in  : BOOL bEnableClosing : FALSE to disable the closing of the modeless window
//                                 TRUE to enable the closing of the modeless window
//     out : 
//     return :
//-----------------------------------------------------------------------------
void CDialogBase::EnableClosingCount(BOOL bEnableClosing)
{
    if (bEnableClosing)
    {
        ::InterlockedDecrement(&this->DisableClosingCount);
#ifdef _DEBUG
        if (this->DisableClosingCount<0)
        {
            BreakInDebugMode;
        }
#endif
    }
    else
        ::InterlockedIncrement(&this->DisableClosingCount);
}

//-----------------------------------------------------------------------------
// Name: CreateButton
// Object: create a button
// Parameters :
//     in  : HWND hWndParent : parent window handle
//           int x : x position relative to parent 
//           int y : y position relative to parent
//           int Width : button width
//           int Height : button height
//           int Identifier : button ID
//           LPCTSTR ButtonText : Button text
//     out : 
//     return : button HWND 
//-----------------------------------------------------------------------------
HWND CDialogBase::CreateButton(
                  int x,
                  int y,
                  int Width,
                  int Height,
                  int Identifier,
                  LPCTSTR ButtonText
                  )
{
    return ::CreateWindowEx( 
                            0,
                            _T("BUTTON"),   // predefined class 
                            ButtonText,       // button text 
                            WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_CENTER | BS_PUSHBUTTON,  // styles 
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
}

//-----------------------------------------------------------------------------
// Name: CreateButton
// Object: create a button
// Parameters :
//     in  : HWND hWndParent : parent window handle
//           int x : x position relative to parent 
//           int y : y position relative to parent
//           int Width : button width
//           int Height : button height
//           int Identifier : button ID
//           LPCTSTR ButtonText : Button text
//     out : 
//     return : button HWND 
//-----------------------------------------------------------------------------

HWND CDialogBase::CreateButtonImage(
                  int x,
                  int y,
                  int Width,
                  int Height,
                  int Identifier,
                  int IconId,int IconWidth,int IconHeight
                  )
{
    HWND hWnd = ::CreateWindowEx( 
                            0,
                            TEXT("BUTTON"),   // predefined class 
                            TEXT(""),       // button text 
                            WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_CENTER | BS_PUSHBUTTON | BS_ICON | BS_FLAT,  // styles BS_FLAT : better rendering on XP
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
    if (hWnd)
        this->SetDlgButtonIconFromHwnd(hWnd,IconId,IconWidth,IconHeight);
    return hWnd;
}


HWND CDialogBase::CreateButtonImageFromhIcon(
                  int x,
                  int y,
                  int Width,
                  int Height,
                  int Identifier,
                  HICON hIcon
                  )
{
    HWND hWnd = ::CreateWindowEx( 
                            0,
                            TEXT("BUTTON"),   // predefined class 
                            TEXT(""),       // button text 
                            WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_CENTER | BS_PUSHBUTTON | BS_ICON,  // styles 
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
    if (hWnd)
        this->SetDlgButtonIconFromHwnd(hWnd,hIcon);
    return hWnd;
}

#if _WIN32_WINNT >= 0x0600
HWND CDialogBase::CreateSplitButton(
                  int x,
                  int y,
                  int Width,
                  int Height,
                  int Identifier,
                  LPCTSTR ButtonText
                  )
{
    return ::CreateWindowEx( 
                            0,
                            _T("BUTTON"),   // predefined class 
                            ButtonText,       // button text 
                            WS_VISIBLE | WS_CHILD | WS_TABSTOP | BS_CENTER | BS_SPLITBUTTON,  // styles 
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
}
#endif
//-----------------------------------------------------------------------------
// Name: CreateStatic
// Object: create a static
// Parameters :
//     in  : HWND hWndParent : parent window handle
//           int x : x position relative to parent 
//           int y : y position relative to parent
//           int Width : static width
//           int Height : static height
//           int Identifier : static ID
//           LPCTSTR StaticText : static text
//     out : 
//     return : static HWND 
//-----------------------------------------------------------------------------
HWND CDialogBase::CreateStatic(
                  int x,
                  int y,
                  int Width,
                  int Height,
                  int Identifier,
                  LPCTSTR StaticText
                  )
{
    return ::CreateWindowEx( 
                            0,
                            _T("STATIC"),   // predefined class 
                            StaticText,       // button text 
                            WS_VISIBLE | WS_CHILD |SS_LEFT,  // styles 
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
}

//-----------------------------------------------------------------------------
// Name: CreateStaticCentered
// Object: create a static with centered align
// Parameters :
//     in  : HWND hWndParent : parent window handle
//           int x : x position relative to parent 
//           int y : y position relative to parent
//           int Width : static width
//           int Height : static height
//           int Identifier : static ID
//           LPCTSTR StaticText : static text
//     out : 
//     return : static HWND 
//-----------------------------------------------------------------------------
HWND CDialogBase::CreateStaticCentered(
                                 int x,
                                 int y,
                                 int Width,
                                 int Height,
                                 int Identifier,
                                 LPCTSTR StaticText
                                 )
{
    return ::CreateWindowEx( 
                            0,
                            _T("STATIC"),   // predefined class 
                            StaticText,       // button text 
                            WS_VISIBLE | WS_CHILD | SS_CENTER,  // styles 
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
}

//-----------------------------------------------------------------------------
// Name: CreateEdit
// Object: create an edit
// Parameters :
//     in  : HWND hWndParent : parent window handle
//           int x : x position relative to parent 
//           int y : y position relative to parent
//           int Width : edit width
//           int Height : edit height
//           int Identifier : edit ID
//           LPCTSTR EditText : edit text
//     out : 
//     return : static HWND 
//-----------------------------------------------------------------------------
HWND CDialogBase::CreateEdit(
                int x,
                int y,
                int Width,
                int Height,
                int Identifier,
                LPCTSTR EditText
                )
{
    return ::CreateWindowEx( 
                            WS_EX_CLIENTEDGE,
                            _T("EDIT"),   // predefined class 
                            EditText,       // button text 
                            WS_VISIBLE | WS_CHILD |WS_TABSTOP | ES_LEFT | ES_AUTOHSCROLL,  // styles 
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
}

//-----------------------------------------------------------------------------
// Name: CreateEditMultiLines
// Object: create a multi lines edit
// Parameters :
//     in  : HWND hWndParent : parent window handle
//           int x : x position relative to parent 
//           int y : y position relative to parent
//           int Width : edit width
//           int Height : edit height
//           int Identifier : edit ID
//           LPCTSTR EditText : edit text
//     out : 
//     return : static HWND 
//-----------------------------------------------------------------------------
HWND CDialogBase::CreateEditMultiLines(
                               int x,
                               int y,
                               int Width,
                               int Height,
                               int Identifier,
                               LPCTSTR EditText
                               )
{
    return ::CreateWindowEx( 
                            WS_EX_CLIENTEDGE,
                            _T("EDIT"),   // predefined class 
                            EditText,       // button text 
                            WS_VISIBLE | WS_CHILD |WS_TABSTOP | ES_LEFT | ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL | ES_WANTRETURN,  // styles 
                            // Size and position values are given explicitly, because 
                            // the CW_USEDEFAULT constant gives zero values for buttons. 
                            x,         // starting x position 
                            y,         // starting y position 
                            Width,        // button width 
                            Height,        // button height 
                            this->hWndDialog,       // parent window 
                            (HMENU)Identifier,       // No menu 
                            this->hInstance, 
                            NULL);      // pointer not needed 
}