/*
Copyright (C) 2004-2020 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2004-2013 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#pragma once

#ifndef _WIN32_WINNT
    // XP or upper
    // #define _WIN32_WINNT 0x0501
    // Vista or upper
    #define _WIN32_WINNT 0x0600 // required for split button (even if in common ctrl 6 and can run under XP with common control 6)
#endif

#include <windows.h>
#pragma warning (push)
#pragma warning(disable : 4005)// for '_stprintf' : macro redefinition in tchar.h
#include <TCHAR.h>
#pragma warning (pop)
#if _WIN32_WINNT >= 0x0600
    #include <CommCtrl.h>
#endif

#ifndef DeleteAndCleanIfNotNull
    // usage : DeleteAndCleanIfNotNull(this->pObject)
    #define DeleteAndCleanIfNotNull(x) if (x){delete x; x=NULL;}
#endif

#ifndef FreeAndCleanIfNotNull
// usage : DeleteAndCleanIfNotNull(this->pObject)
#define FreeAndCleanIfNotNull(x) if (x){free(x); x=NULL;}
#endif

#ifndef ReleaseAndCleanIfNotNull
    // usage : ReleaseAndCleanIfNotNull(this->pObject)
    #define ReleaseAndCleanIfNotNull(x) if (x){x->Release(); x=NULL;}
#endif

#ifndef DeleteAndCleanArrayIfNotNull
    // usage : DeleteAndCleanArrayIfNotNull(Array)
    #define DeleteAndCleanArrayIfNotNull(Array) if (Array){delete[] Array; Array=NULL;}
#endif

#ifndef CloseHandleAndCleanIfNotNull
    // usage : CloseHandleAndCleanIfNotNull(hHandle)
    #define CloseHandleAndCleanIfNotNull(x) if (x){::CloseHandle(x); x=NULL;}
#endif

#ifndef DestroyIconAndCleanIfNotNull
    // usage : DestroyIconAndCleanIfNotNull(hIcon)
    #define DestroyIconAndCleanIfNotNull(x) if (x){::DestroyIcon(x); x=NULL;}
#endif

#ifndef DestroyWindowAndCleanIfNotNull
    // usage : DestroyIconAndCleanIfNotNull(hWnd)
    #define DestroyWindowAndCleanIfNotNull(x) if (x){::DestroyWindow(x); x=NULL;}
#endif

#ifndef DeleteFontAndCleanIfNotNull
    // usage : DeleteFontAndCleanIfNotNull(hFont)
    #define DeleteFontAndCleanIfNotNull(x) if (x){DeleteObject((HGDIOBJ)(HFONT)(x)); x=NULL;}
#endif

#ifndef _DefineStringLen
    // _DefineStringLen : remove string \0 from array size
    #define _DefineStringLen(Array) (_countof(Array)-1)
#endif

#ifndef _DefineStringLenWithEndingZero
    #define _DefineStringLenWithEndingZero(Array) (_countof(Array))
#endif

#ifndef BreakInDebugMode
    #ifdef _DEBUG
        #define BreakInDebugMode if ( ::IsDebuggerPresent() ) { ::DebugBreak();}
    #else 
        #define BreakInDebugMode 
    #endif
#endif

#define CDialogBase_SpaceBetweenControls 5

#define CDialogBase_SHELL_DLL _T("shell32.dll")

#include "DialogModelessProtection.h"

class CDialogBase
{
protected:
    HWND hWndDialog;
    HWND hWndParentDialog;
    HINSTANCE hInstance;
    BOOL bInsideDialogBoxFunction;
    int IconResourceId;
    HCURSOR hWaitCursor;
    BOOL bWaitCursor;
    BOOL bDragAcceptFileRequired; // if we have to enable drag accept file
    BOOL bDragAcceptFileStatusEnabled;// if drag accept file is already enabled
    BOOL bInitDone;
    BOOL bModalDialog;
    volatile LONG DisableClosingCount;
    BOOL bCloseModalSubDialogsIfAny;

    CRITICAL_SECTION CriticalSection;

    static INT_PTR CALLBACK DlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

    typedef struct tagMoveGroupStruct
    {
        POINT  Point;
        CDialogBase* pDialogBase;
    }MOVE_GROUP_STRUCT,*PMOVE_GROUP_STRUCT;

    static BOOL CallBackRedrawGroup(HWND Item,PVOID UserParam);
    static BOOL CallBackShowGroup(HWND Item,PVOID UserParam);
    static BOOL CallBackEnableGroup(HWND Item,PVOID UserParam);
    static BOOL CallBackMoveGroup(HWND Item,PVOID UserParam);
    typedef BOOL (*pfCallBackGroupAction)(HWND Item,PVOID UserParam);

    static BOOL CALLBACK EnumWndProc_CloseModalSubDialogsIfAny(HWND hWnd,LPARAM lParam);
    static BOOL CALLBACK EnumWndProc_GetTopMostModalPopUpSubDialogs(HWND hWnd,LPARAM lParam);

    BOOL ParseGroupForAction(HWND hWndGroup,pfCallBackGroupAction CallBackFunction,PVOID UserParam);
    BOOL ScreenToClient(IN OUT LPRECT pRect);
    virtual BOOL OnInitInternal(){return this->OnInit();}
    virtual BOOL OnCloseInternal()
    {
        if (this->DisableClosingCount>0)
        {
            this->Flash();
            BreakInDebugMode;
            return FALSE;
        }
        return this->OnClose();
    }

    BOOL CloseInternal(INT_PTR Code,BOOL bCloseModalSubDialogsIfAny,BOOL bForceClosing);
    virtual BOOL CloseModalSubDialogsIfAny();
public:
    CDialogBase(void);
    virtual ~CDialogBase(void); // MUST BE virtual
    void Flash();
    static CDialogBase* GetAssociatedObject(HWND hWndDialog);
    FORCEINLINE BOOL HasModalSubWindowDisplayed(){return (this->DisableClosingCount /* !=0 */);}
    FORCEINLINE BOOL IsInsideDialogBoxFunction(){return this->bInsideDialogBoxFunction;}
    FORCEINLINE BOOL IsInitialized(){return this->bInitDone;}// return TRUE after OnInit() completion
    virtual BOOL OnInit(){return TRUE;}//return TRUE to direct the system to set the keyboard focus to the control specified by wParam. Otherwise, it should return FALSE to prevent the system from setting the default keyboard focus
    virtual BOOL OnClose(){return TRUE;}//return FALSE to cancel closing

    virtual void OnCommand(WPARAM wParam,LPARAM lParam)
    {
        UNREFERENCED_PARAMETER(lParam);
        switch(LOWORD(wParam))
        {
            case IDOK:
                // can't be done, else if not overwritten, dialog close when user hits the return key
                // this->Close(IDOK);
                break;
            case IDCANCEL:
                this->Close(IDCANCEL);
                break;
        }
    }
    virtual BOOL OnNotify(WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);return FALSE;}
    virtual void OnDropFiles(WPARAM wParam,LPARAM lParam)
    {
        UNREFERENCED_PARAMETER(wParam);
        UNREFERENCED_PARAMETER(lParam);
        /*
        HDROP hDrop = (HDROP)wParam;
        TCHAR szFileName[MAX_PATH];
    
        // retrieve dialog subitem window handle (if required to do subitem specific action)
        HWND SubWindowHandle;
        POINT pt;
        ::DragQueryPoint(hDrop, &pt);
        ::ClientToScreen(this->hWndDialog,&pt);
        SubWindowHandle=::WindowFromPoint(pt);
        int CtrlId = ::GetDlgCtrlID(SubWindowHandle);

        // 1) if we just support one name
        // get first file name
        ::DragQueryFile(hDrop, 0,szFileName, MAX_PATH);
        // DoSomeThing(szFileName);

        // 2) if we support multiple files
        // get number of files count in the drag and drop (if required)
        UINT NbFiles;
        NbFiles=::DragQueryFile(hDrop, 0xFFFFFFFF,NULL, MAX_PATH);

        // load all files (if required)
        UINT Cnt;
        for (Cnt=0;Cnt<NbFiles;Cnt++)
        {
            ::DragQueryFile(hDrop, Cnt,szFileName, MAX_PATH);
            // DoSomeThing(szFileName);
        }

        ::DragFinish(hDrop);
        */
    }
    virtual void OnMouseDown(WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);}
    virtual void OnMouseUp(WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);}
    virtual void OnMouseMove(WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);}
    virtual void OnSize(WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);}
    virtual void OnSizing(WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);}
    virtual void OnActivate(WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);}

    // Notice: OnMessage can't be used if message should return FALSE (like WM_QUERYENDSESSION)
    // as DlgProc returning FALSE are for not handled message
    // in those cases you need to add subclassing with SetWindowSubclass / RemoveWindowSubclass
    virtual LRESULT OnMessage(UINT uMsg,WPARAM wParam,LPARAM lParam){UNREFERENCED_PARAMETER(uMsg);UNREFERENCED_PARAMETER(wParam);UNREFERENCED_PARAMETER(lParam);return FALSE;}

    virtual HWND CreateModless(HINSTANCE hInstance,HWND hWndParentDialog,int ResourceId,int IconResourceId);
    virtual HWND CreateModlessIndirect(HINSTANCE hInstance,HWND hWndParentDialog,LPCDLGTEMPLATE hDialogTemplate,int IconResourceId);
    virtual INT_PTR Show(HINSTANCE hInstance,HWND hWndParentDialog,int ResourceId,int IconResourceId);
    virtual INT_PTR ShowIndirect(HINSTANCE hInstance,HWND hWndParentDialog,LPCDLGTEMPLATE hDialogTemplate,int IconResourceId);
    // virtual BOOL Close(INT_PTR Code = 0);
    virtual BOOL Close(){return this->Close(0);}
    virtual BOOL Close(INT_PTR Code);
    virtual BOOL CloseEx(INT_PTR Code,BOOL bCloseModalSubDialogsIfAny);
    virtual BOOL CloseFromOtherThread(BOOL bCloseModalSubDialogsIfAny,BOOL bWaitEndOfDialog,DWORD WaitTimeout);
    virtual BOOL WaitDialogEnd(DWORD TimeOutInMs); 
    virtual BOOL WaitDialogThreadEnd(DWORD TimeOutInMs);

    FORCEINLINE HWND GetControlHandle(){return this->hWndDialog;}
    FORCEINLINE BOOL GetInternalRect(LPRECT lpRect){return ::GetClientRect(this->hWndDialog,lpRect);}
    FORCEINLINE BOOL GetWindowRect(LPRECT lpRect){return ::GetWindowRect(this->hWndDialog,lpRect);}
    DWORD GetHeight();
    DWORD GetWidth();
    DWORD GetInternalHeight();
    DWORD GetInternalWidth();
    FORCEINLINE HINSTANCE GetInstance(){return this->hInstance;}
    FORCEINLINE BOOL IsModal(){return this->bModalDialog;}
    FORCEINLINE HWND GetDlgItem(IN int nIDDlgItem){return ::GetDlgItem(this->hWndDialog,nIDDlgItem);}
    FORCEINLINE int GetControlId(IN HWND hWindow){return ::GetDlgCtrlID(hWindow);}
    FORCEINLINE LRESULT SendDlgMessage(IN int nIDDlgItem,IN UINT message,IN WPARAM wParam,IN LPARAM lParam){return ::SendMessage(::GetDlgItem(this->hWndDialog,nIDDlgItem),message,wParam,lParam);}
    FORCEINLINE LRESULT PostDlgMessage(IN int nIDDlgItem,IN UINT message,IN WPARAM wParam,IN LPARAM lParam){return ::PostMessage(::GetDlgItem(this->hWndDialog,nIDDlgItem),message,wParam,lParam);}

    FORCEINLINE LRESULT SendDlgMessage(IN UINT message,IN WPARAM wParam,IN LPARAM lParam){return ::SendMessage(this->hWndDialog,message,wParam,lParam);}
    FORCEINLINE LRESULT PostDlgMessage(IN UINT message,IN WPARAM wParam,IN LPARAM lParam){return ::PostMessage(this->hWndDialog,message,wParam,lParam);}

    // avoid macro definition (else we have to refactor all projects using current class,
    // We have to use SetDlgItemTextT and GetDlgItemTextT, as 
    //      SetDlgItemText is define as SetDlgItemTextA or SetDlgItemTextW
    //      GetDlgItemText is define as GetDlgItemTextA or GetDlgItemTextW

    FORCEINLINE BOOL SetDlgItemInt(IN int nIDDlgItem,IN UINT uValue,IN BOOL bSigned){return ::SetDlgItemInt(this->hWndDialog, nIDDlgItem, uValue, bSigned);}
    FORCEINLINE UINT GetDlgItemInt(IN int nIDDlgItem,OUT BOOL* lpTranslated,IN BOOL bSigned){return ::GetDlgItemInt(this->hWndDialog, nIDDlgItem, lpTranslated, bSigned);}
    FORCEINLINE BOOL SetDlgItemTextT(IN int nIDDlgItem,LPCTSTR lpszString){return ::SetWindowText(::GetDlgItem(this->hWndDialog,nIDDlgItem), lpszString);}
    FORCEINLINE BOOL SetDlgItemTextA(IN int nIDDlgItem,LPCSTR lpszString){return ::SetWindowTextA(::GetDlgItem(this->hWndDialog,nIDDlgItem), lpszString);}
    FORCEINLINE BOOL SetDlgItemTextW(IN int nIDDlgItem,LPCWSTR lpszString){return ::SetWindowTextW(::GetDlgItem(this->hWndDialog,nIDDlgItem), lpszString);}

    // nMaxCount : Specifies the maximum number of TCHARs to be copied, including the terminating null character
    FORCEINLINE int GetDlgItemTextT(IN int nIDDlgItem,OUT LPTSTR lpString,IN int nMaxCount){*lpString=0;return ::GetWindowText(::GetDlgItem(this->hWndDialog,nIDDlgItem), lpString,nMaxCount);}
    FORCEINLINE int GetDlgItemTextA(IN int nIDDlgItem,OUT LPSTR lpString,IN int nMaxCount){*lpString=0;return ::GetWindowTextA(::GetDlgItem(this->hWndDialog,nIDDlgItem), lpString,nMaxCount);}
    FORCEINLINE int GetDlgItemTextW(IN int nIDDlgItem,OUT LPWSTR lpString,IN int nMaxCount){*lpString=0;return ::GetWindowTextW(::GetDlgItem(this->hWndDialog,nIDDlgItem), lpString,nMaxCount);}
    // The return value is the length of the text in TCHARs, not including the terminating null character
    FORCEINLINE int GetDlgItemTextLength(IN int nIDDlgItem){return ::GetWindowTextLength(::GetDlgItem(this->hWndDialog,nIDDlgItem));}

    FORCEINLINE BOOL SetEditReadOnly(IN int EditID,BOOL bReadOnly){return (BOOL)::SendMessage(::GetDlgItem(this->hWndDialog,EditID),EM_SETREADONLY,(WPARAM)bReadOnly,0);}
    // SetEditMaxChar maximum number of TCHARs the user can enter. This number does not include the null terminator
    //                MaxNbChar depends of os version and edit type (see EM_SETLIMITTEXT documentation)
    FORCEINLINE BOOL SetEditMaxChar(IN int EditID,DWORD MaxNbChar){return (BOOL)::SendMessage(::GetDlgItem(this->hWndDialog,EditID),EM_SETLIMITTEXT,(WPARAM)MaxNbChar,0);}

    FORCEINLINE BOOL IsDlgButtonChecked(IN int nIDDlgItem){return (::SendMessage(::GetDlgItem(this->hWndDialog,nIDDlgItem),BM_GETCHECK,0,0)!=BST_UNCHECKED);}
    FORCEINLINE LRESULT SetDlgButtonCheckState(IN int nIDDlgItem,IN BOOL bChecked){return ::SendMessage(::GetDlgItem(this->hWndDialog,nIDDlgItem),BM_SETCHECK,bChecked ? BST_CHECKED : BST_UNCHECKED,0);}
    FORCEINLINE BOOL EnableDlgItem(IN int nIDDlgItem,IN BOOL bEnable){return ::EnableWindow(::GetDlgItem(this->hWndDialog,nIDDlgItem),bEnable);}
    FORCEINLINE BOOL EnableDlgItemFromHwnd(IN HWND hWnd,IN BOOL bEnable){return ::EnableWindow( hWnd,bEnable);}
    FORCEINLINE BOOL IsDlgItemEnabled(IN int nIDDlgItem){return ::IsWindowEnabled(::GetDlgItem(this->hWndDialog,nIDDlgItem));}

    // do not call Destroy icon for returned icon handle
    FORCEINLINE HICON LoadIcon(int IconId,int Width,int Height){return (HICON)::LoadImage(this->hInstance,MAKEINTRESOURCE(IconId),IMAGE_ICON,Width,Height,LR_DEFAULTCOLOR|LR_SHARED);}
    FORCEINLINE LRESULT SetStaticIcon(int StaticControlId,HICON hIcon){return ::SendMessage(::GetDlgItem(this->hWndDialog,StaticControlId),STM_SETICON,(WPARAM)hIcon,0);}
    HICON SetStaticIcon(int StaticControlId,int IconId,int Width,int Height);
    FORCEINLINE LRESULT SetDlgButtonIcon(int ButtonControlId,HICON hIcon){return ::SendMessage(::GetDlgItem(this->hWndDialog,ButtonControlId),BM_SETIMAGE,(WPARAM)IMAGE_ICON,(LPARAM)hIcon);}
    FORCEINLINE LRESULT SetDlgButtonIconFromHwnd(HWND hWnd,HICON hIcon){return ::SendMessage(hWnd,BM_SETIMAGE,(WPARAM)IMAGE_ICON,(LPARAM)hIcon);}
    HICON SetDlgButtonIcon(int ButtonControlId,int IconId,int Width,int Height){return this->SetDlgButtonIconFromHwnd(::GetDlgItem(this->hWndDialog,ButtonControlId),IconId, Width, Height);}
    HICON SetDlgButtonIconFromHwnd(HWND hWnd,int IconId,int Width,int Height);
    
    void SetDlgIcon(HINSTANCE hInstance,int IconResourceId);
    FORCEINLINE BOOL SetCaption(TCHAR* DlgCaption){return ::SetWindowText(this->hWndDialog, DlgCaption);}
    FORCEINLINE int GetCaption(TCHAR* DlgCaption,int DlgCaptionMaxSize){return ::GetWindowText(this->hWndDialog, DlgCaption,DlgCaptionMaxSize);}
    void SetWaitCursor(BOOL bWaitCursor);
    
    void CenterOwner();
    void CenterOnWindow(HWND hWnd);

    FORCEINLINE BOOL Hide(){return ::ShowWindow(this->hWndDialog,SW_HIDE);}
    FORCEINLINE BOOL Hide(IN int nIDDlgItem){return ::ShowWindow(this->GetDlgItem(nIDDlgItem),SW_HIDE);}
    FORCEINLINE BOOL Hide(IN HWND hWnd){return ::ShowWindow(hWnd,SW_HIDE);}

    FORCEINLINE BOOL Show(){return ::ShowWindow(this->hWndDialog,SW_SHOWNORMAL);}
    FORCEINLINE BOOL Show(IN int nIDDlgItem){return ::ShowWindow(this->GetDlgItem(nIDDlgItem),SW_SHOWNORMAL);}
    FORCEINLINE BOOL Show(IN int nIDDlgItem,IN int CmdShow){return ::ShowWindow(this->GetDlgItem(nIDDlgItem),CmdShow);}
    FORCEINLINE BOOL Show(IN HWND hWnd){return ::ShowWindow(hWnd,SW_SHOWNORMAL);}

    FORCEINLINE BOOL IsVisible(){return ::IsWindowVisible(this->hWndDialog);}
    FORCEINLINE BOOL IsVisible(IN HWND hWnd){return ::IsWindowVisible(hWnd);}
    FORCEINLINE BOOL IsDlgItemVisible(IN int nIDDlgItem){return ::IsWindowVisible(this->GetDlgItem(nIDDlgItem));}

    void BringToTop();
    FORCEINLINE BOOL SetForeground(){return ::SetForegroundWindow(this->hWndDialog);}
    FORCEINLINE BOOL SetFocus(HWND hWndControl){return ::PostMessage(this->hWndDialog, WM_NEXTDLGCTL, (WPARAM)hWndControl, TRUE);}// we must use WM_NEXTDLGCTL for a dialog proc

    FORCEINLINE HWND GetParentDialog(){return (HWND)::GetWindowLongPtr(this->hWndDialog,GWLP_HWNDPARENT);}
    HWND GetParentPopUpDialog();
    static HWND GetParentPopUpDialog(HWND hWnd);
    static HWND GetParentTopMostPopUpDialog(HWND hWnd);

    // get window thread and pid from handle
    FORCEINLINE DWORD GetWindowThreadID(){return ::GetWindowThreadProcessId(this->hWndDialog,NULL);}
    FORCEINLINE DWORD GetWindowProcessID(){DWORD ProcessId = NULL; ::GetWindowThreadProcessId(this->hWndDialog,&ProcessId);return ProcessId;}

    FORCEINLINE LONG_PTR SetStyle(IN LONG_PTR NewStyle){return ::SetWindowLongPtr(this->hWndDialog,GWL_STYLE,NewStyle);}
    FORCEINLINE LONG_PTR SetStyle(IN int nIDDlgItem,IN LONG_PTR NewStyle){return ::SetWindowLongPtr(this->GetDlgItem(nIDDlgItem),GWL_STYLE,NewStyle);}
    FORCEINLINE LONG_PTR SetStyleFromHwnd(IN HWND hItem,IN LONG_PTR NewStyle){return ::SetWindowLongPtr(hItem,GWL_STYLE,NewStyle);}

    FORCEINLINE LONG_PTR GetStyle(){return ::GetWindowLongPtr(this->hWndDialog,GWL_STYLE);}
    FORCEINLINE LONG_PTR GetStyle(IN int nIDDlgItem){return ::GetWindowLongPtr(this->GetDlgItem(nIDDlgItem),GWL_STYLE);}
    FORCEINLINE LONG_PTR GetStyleFromHwnd(IN HWND hItem){return ::GetWindowLongPtr(hItem,GWL_STYLE);}

    FORCEINLINE LONG_PTR SetStyleEx(IN LONG_PTR NewStyle){return ::SetWindowLongPtr(this->hWndDialog,GWL_EXSTYLE,NewStyle);}
    FORCEINLINE LONG_PTR SetStyleEx(IN int nIDDlgItem,IN LONG_PTR NewStyle){return ::SetWindowLongPtr(this->GetDlgItem(nIDDlgItem),GWL_EXSTYLE,NewStyle);}
    FORCEINLINE LONG_PTR SetStyleExFromHwnd(IN HWND hItem,IN LONG_PTR NewStyle){return ::SetWindowLongPtr(hItem,GWL_EXSTYLE,NewStyle);}

    FORCEINLINE LONG_PTR GetStyleEx(){return ::GetWindowLongPtr(this->hWndDialog,GWL_EXSTYLE);}
    FORCEINLINE LONG_PTR GetStyleEx(IN int nIDDlgItem){return ::GetWindowLongPtr(this->GetDlgItem(nIDDlgItem),GWL_EXSTYLE);}
    FORCEINLINE LONG_PTR GetStyleExFromHwnd(IN HWND hItem){return ::GetWindowLongPtr(hItem,GWL_EXSTYLE);}

    BOOL EnableDragAndDrop(BOOL bEnable);

    BOOL EnableRedraw(IN BOOL bEnable);
    BOOL EnableRedraw(IN int nIDDlgItem,IN BOOL bEnable);
    BOOL EnableRedrawFromHwnd(IN HWND hWnd,IN BOOL bEnable);

    BOOL Redraw();
    BOOL Redraw(IN int nIDDlgItem);
    BOOL RedrawFromHwnd(HWND hItem);
    BOOL Invalidate(IN int nIDDlgItem);
    BOOL InvalidateFromHwnd(HWND hItem);
    FORCEINLINE BOOL Invalidate(){return this->InvalidateFromHwnd(this->hWndDialog);}
    BOOL GetClientWindowRect(IN int nIDDlgItem,OUT LPRECT lpRect);
    BOOL GetClientWindowRectFromHwnd(IN HWND hItem,OUT LPRECT lpRect);

    FORCEINLINE BOOL SetPosition(LONG x,LONG y){return ::SetWindowPos(this->hWndDialog,0,x,y,0,0,SWP_NOZORDER | SWP_NOSIZE | SWP_NOACTIVATE);}
    FORCEINLINE BOOL SetPositionFromHwnd(HWND hWnd,LONG x,LONG y){return ::SetWindowPos(hWnd,0,x,y,0,0,SWP_NOZORDER | SWP_NOSIZE | SWP_NOACTIVATE);}
    FORCEINLINE BOOL SetPositionFromId(int Id,LONG x,LONG y){return ::SetWindowPos(this->GetDlgItem(Id),0,x,y,0,0,SWP_NOZORDER | SWP_NOSIZE | SWP_NOACTIVATE);}
    BOOL OffsetPositionFromHwnd(HWND hWnd,LONG xOffset,LONG yOffset);
    BOOL OffsetPositionFromId(int Id,LONG xOffset,LONG yOffset);
    FORCEINLINE BOOL SetRect(RECT* pRect){return ::SetWindowPos(this->hWndDialog,NULL,pRect->left,pRect->top,pRect->right-pRect->left,pRect->bottom-pRect->top,SWP_NOACTIVATE | SWP_NOZORDER);}
    FORCEINLINE BOOL SetRectFromHwnd(HWND hWnd,RECT* pRect){return ::SetWindowPos(hWnd,NULL,pRect->left,pRect->top,pRect->right-pRect->left,pRect->bottom-pRect->top,SWP_NOACTIVATE | SWP_NOZORDER);}
    FORCEINLINE BOOL SetRectFromId(int Id,RECT* pRect){return ::SetWindowPos(this->GetDlgItem(Id),NULL,pRect->left,pRect->top,pRect->right-pRect->left,pRect->bottom-pRect->top,SWP_NOACTIVATE | SWP_NOZORDER);}
    FORCEINLINE BOOL SetSize(LONG Width,LONG Height){return ::SetWindowPos(this->hWndDialog,NULL,0,0,Width,Height,SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOMOVE);}
    FORCEINLINE BOOL SetSizeFromHwnd(HWND hWnd,LONG Width,LONG Height){return ::SetWindowPos(hWnd,NULL,0,0,Width,Height,SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOMOVE);}
    FORCEINLINE BOOL SetSizeFromId(int Id,LONG Width,LONG Height){return ::SetWindowPos(this->GetDlgItem(Id),NULL,0,0,Width,Height,SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOMOVE);}


    void DialogUnitToPixels(RECT* pRect);
    void PixelsToDialogUnit(RECT* pRect);

    HWND GetTopMostModalPopUpSubDialogs();

    ////////////////
    // Combo box
    ////////////////
    HWND GetComboboxEditHandle(IN int nIDDlgItem);
    BOOL IsComboboxHandle(IN int nIDDlgItem,IN HWND hItemToTest);

    /////////////////
    // group actions
    /////////////////
    BOOL RedrawGroup(IN int nIDDlgItem);
    BOOL RedrawGroupFromHwnd(IN HWND hWndGroup);
    BOOL ShowGroup(IN int nIDDlgItem,IN int CmdShow);
    BOOL ShowGroupFromHandle(IN HWND hWndGroup,IN int CmdShow);
    BOOL EnableGroup(IN int nIDDlgItem,IN BOOL bEnable);
    BOOL EnableGroupFromHandle(IN HWND hWndGroup,IN BOOL bEnable);
    BOOL MoveGroupTo(IN int nIDDlgItem,IN POINT* pPoint);
    BOOL MoveGroupToFromHandle(IN HWND hWndGroup,IN POINT* pPoint);
    BOOL OffsetGroupPosition(IN int nIDDlgItem,IN LONG xOffset,IN LONG yOffset);
    BOOL OffsetGroupPositionFromHandle(IN HWND hWndGroup,IN LONG xOffset,IN LONG yOffset);

    //////////////////
    // font
    //////////////////
    HFONT GetFont(){return this->GetFontFromHandle(this->hWndDialog);}
    HFONT GetFontFromHandle(IN HWND hWnd){return (HFONT)::SendMessage(hWnd,WM_GETFONT,0,0);}
    HFONT GetFont(IN int nIDDlgItem){return this->GetFontFromHandle(this->GetDlgItem(nIDDlgItem));}
    void SetFontFromHandle(IN HWND hWnd,IN HFONT hFont){::SendMessage(hWnd,WM_SETFONT,(WPARAM)hFont,TRUE);}
    void SetFont(IN int nIDDlgItem,IN HFONT hFont){this->SetFontFromHandle(this->GetDlgItem(nIDDlgItem),hFont);}

    //////////////////
    // modeless specific function
    //////////////////
    // Note: if a modeless dialog has a child modal dialog, the child modal loop will handle the main dialog event 
    // this is a problem when main dialog own modeless dialog but is not parent
    // the close event of the main dialog is manage by the modeless child message loop,
    // so main dialog and modeless and modeless child are destroyed when the message processing ends,
    // we return to modeless method but the modeless object is destroyed --> Crash
    // to solve this the easiest way is to disable main dialog when showing a modal/messageBox dialog inside a modeless dialog
    // trouble EnableWindow is not counted... not possible if multiple modeless window (or need of mutex is required)
    //::EnableWindow(::GetAncestor(this->hWndDialog, GA_ROOT),FALSE);
    //::EnableWindow(::GetAncestor(this->hWndDialog, GA_ROOTOWNER),FALSE);
    void EnableClosingCount(BOOL bEnableClosing);

    //////////////////
    // message report
    //////////////////
    enum ReportMessageTypes
    {
        ReportMessageType_ERROR,
        ReportMessageType_ERROR_QUESTION_YES_NO,
        ReportMessageType_WARNING,
        ReportMessageType_WARNING_QUESTION_YES_NO,
        ReportMessageType_INFO,
        ReportMessageType_INFO_QUESTION_YES_NO,
        ReportMessageType_QUESTION_YES_NO,
        ReportMessageType_QUESTION_YES_NO_CANCEL
    };
    virtual int ReportMessage(TCHAR* pszMsg, ReportMessageTypes MsgType);
    virtual int ReportMessageEx(TCHAR* pszMsg, TCHAR* OptCaption, ReportMessageTypes MsgType);// OptCaption can be NULL
    virtual int ReportError(TCHAR* pszMsg);

    //////////////////
    // control creation
    //////////////////
    virtual HWND CreateButton(int x,int y,int Width,int Height,int Identifier,LPCTSTR ButtonText);
    virtual HWND CreateButtonImage(int x,int y,int Width,int Height,int Identifier,int IconId,int IconWidth,int IconHeight);
    virtual HWND CreateButtonImageFromhIcon(int x,int y,int Width,int Height,int Identifier,HICON hIcon);
#if _WIN32_WINNT >= 0x0600
    virtual HWND CreateSplitButton(int x,int y,int Width,int Height,int Identifier,LPCTSTR ButtonText);
#endif
    virtual HWND CreateStatic(int x,int y,int Width,int Height,int Identifier,LPCTSTR StaticText);
    virtual HWND CreateStaticCentered(int x,int y,int Width,int Height,int Identifier,LPCTSTR StaticText);
    virtual HWND CreateEdit(int x,int y,int Width,int Height,int Identifier,LPCTSTR EditText);
    virtual HWND CreateEditMultiLines(int x,int y,int Width,int Height,int Identifier,LPCTSTR EditText);

protected:
    void GetDefaultCaptionAndFlagFromReportMessageType(IN ReportMessageTypes MsgType,OUT const TCHAR** ppszCaption,OUT UINT* pMessageBoxFlag);
};
